"""Classes and methods for working with deterministic finite automata."""

from __future__ import annotations

import array
from collections import defaultdict, deque
from itertools import chain, count
from random import Random
from typing import (
    AbstractSet,
    Any,
    Callable,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Generator,
    Iterator,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    cast,
)

import networkx as nx
from cached_method import cached_method
from typing_extensions import Self, TypeAlias

import automata.base.exceptions as exceptions
import automata.fa.fa as fa
import automata.fa.nfa as nfa
from automata.base.utils import (
    PartitionRefinement,
    get_reachable_nodes,
    get_renaming_function,
    pairwise,
)

DFAStateT = fa.FAStateT

DFASymbolT = str
DFAPathT = Mapping[DFASymbolT, DFAStateT]
DFATransitionsT = Mapping[DFAStateT, DFAPathT]

ExpandStateReturnType = Iterator[Tuple[DFASymbolT, DFAStateT]]
ExpandStateFn = Callable[[DFAStateT], ExpandStateReturnType]
IsFinalStateFn = Callable[[DFAStateT], bool]
TargetStateFn = Callable[[DFAStateT], bool]


class DFA(fa.FA):
    """
    The `DFA` class is a subclass of `FA` and represents a deterministic finite
    automaton.

    Every DFA can be rendered natively inside of a Jupyter notebook
    (automatically calling `show_diagram` without any arguments) if installed
    with the `visual` optional dependency.

    Parameters
    ----------
    states : AbstractSet[DFAStateT]
        Set of the DFA's valid states.
    input_symbols : AbstractSet[str]
        Set of the DFA's valid input symbols, each of which is a singleton
        string
    transitions : Mapping[DFAStateT, Mapping[str, DFAStateT]]
        Dict consisting of the transitions for each state. Each key is a
        state name, and each value is another dict which maps a symbol
        (the key) to a state (the value).
    initial_state : DFAStateT
        The initial state for this DFA.
    final_states : AbstractSet[DFAStateT]
        A set of final states for this DFA
    allow_partial : bool, default: True
        By default, each DFA state must have a transition to
        every input symbol; if this parameter is `True`, you can disable this
        characteristic (such that any DFA state can have fewer transitions than input
        symbols). Note that a DFA must always have every state represented in the
        transition dictionary, even if there are no transitions on input symbols
        leaving a state (dictionary is left empty in that case).
    """

    __slots__ = (
        "states",
        "input_symbols",
        "transitions",
        "initial_state",
        "final_states",
        "allow_partial",
        "_count_cache",
        "_word_cache",
        # These two entries are to allow for caching methods
        "__dict__",
        "__weakref__",
    )

    allow_partial: bool
    _word_cache: List[DefaultDict[DFAStateT, List[str]]]
    _count_cache: List[DefaultDict[DFAStateT, int]]

    def __init__(
        self,
        *,
        states: AbstractSet[DFAStateT],
        input_symbols: AbstractSet[str],
        transitions: DFATransitionsT,
        initial_state: DFAStateT,
        final_states: AbstractSet[DFAStateT],
        allow_partial: bool = False,
    ) -> None:
        """Initialize a complete DFA."""
        super().__init__(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state,
            final_states=final_states,
            allow_partial=allow_partial,
        )

        self.clear_cache()

    def clear_cache(self) -> None:
        """
        Resets the word and count caches.
        Can be called if too much memory is being used.
        """
        object.__setattr__(self, "_word_cache", [])
        object.__setattr__(self, "_count_cache", [])

    def __eq__(self, other: Any) -> bool:
        """
        Return True if two DFAs are equivalent. Uses an optimized version of
        the Hopcroft-Karp algorithm. See https://arxiv.org/abs/0907.5058
        """

        DFAStatePairT = Tuple[DFAStateT, int]

        # Must be another DFA and have equal alphabets
        if not isinstance(other, DFA) or self.input_symbols != other.input_symbols:
            return NotImplemented

        operand_dfas = (self, other)
        initial_state_a = (self.initial_state, 0)
        initial_state_b = (other.initial_state, 1)

        def is_final_state(state_pair: DFAStatePairT) -> bool:
            state, operand_index = state_pair
            return state in operand_dfas[operand_index].final_states

        def transition(state_pair: DFAStatePairT, symbol: str) -> DFAStatePairT:
            state, operand_index = state_pair
            return (
                operand_dfas[operand_index]._get_next_current_state(state, symbol),
                operand_index,
            )

        # Get data structures
        state_sets = nx.utils.union_find.UnionFind((initial_state_a, initial_state_b))
        pair_stack: Deque[Tuple[DFAStatePairT, DFAStatePairT]] = deque()

        # Do union find
        state_sets.union(initial_state_a, initial_state_b)
        pair_stack.append((initial_state_a, initial_state_b))

        while pair_stack:
            q_a, q_b = pair_stack.pop()

            if is_final_state(q_a) ^ is_final_state(q_b):
                return False

            for symbol in self.input_symbols:
                r_1 = state_sets[transition(q_a, symbol)]
                r_2 = state_sets[transition(q_b, symbol)]

                if r_1 != r_2:
                    state_sets.union(r_1, r_2)
                    pair_stack.append((r_1, r_2))

        return True

    def __le__(self, other: DFA) -> bool:
        """Return True if this DFA is a subset of (or equal to) another DFA."""
        if isinstance(other, DFA):
            return self.issubset(other)
        else:
            return NotImplemented

    def __ge__(self, other: DFA) -> bool:
        """Return True if this DFA is a superset of another DFA."""
        if isinstance(other, DFA):
            return self.issuperset(other)
        else:
            return NotImplemented

    def __lt__(self, other: DFA) -> bool:
        """Return True if this DFA is a strict subset of another DFA."""
        if isinstance(other, DFA):
            return self <= other and self != other
        else:
            return NotImplemented

    def __gt__(self, other: DFA) -> bool:
        """Return True if this DFA is a strict superset of another DFA."""
        if isinstance(other, DFA):
            return self >= other and self != other
        else:
            return NotImplemented

    def __sub__(self, other: DFA) -> Self:
        """Return a DFA that is the difference of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.difference(other)
        else:
            return NotImplemented

    def __or__(self, other: DFA) -> Self:
        """Return the union of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.union(other)
        else:
            return NotImplemented

    def __and__(self, other: DFA) -> Self:
        """Return the intersection of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.intersection(other)
        else:
            return NotImplemented

    def __xor__(self, other: DFA) -> Self:
        """Return the symmetric difference of this DFA and another DFA."""
        if isinstance(other, DFA):
            return self.symmetric_difference(other)
        else:
            return NotImplemented

    def __invert__(self) -> Self:
        """Return the complement of this DFA and another DFA."""
        return self.complement()

    def __iter__(self) -> Iterator[str]:
        """
        Iterates through all words in the language represented by the DFA. The
        words are ordered first by length and then by the order of the input
        symbol set.
        """
        i = self.minimum_word_length()
        limit = self.maximum_word_length()
        while limit is None or i <= limit:
            yield from self.words_of_length(i)
            i += 1

    def __len__(self) -> int:
        """Returns the cardinality of the language represented by the DFA."""
        return self.cardinality()

    def to_partial(self, *, retain_names: bool = False, minify: bool = True) -> Self:
        """
        Turns a DFA (complete or not) into a partial DFA.
        Removes dead states and trap states (except the initial state)
            and all edges leading to them.

        Parameters
        ----------
        minify : bool, default: True
            Whether to perform a minify operation while converting to
            a partial DFA.
        retain_names : bool, default: True
            Whether to retain state names during minification.

        Returns
        -------
        Self
            An equivalent partial DFA.
        """

        # Still want to try and remove dead states in the case of a partial DFA
        graph = self._get_digraph()
        live_states = get_reachable_nodes(graph, [self.initial_state])
        non_trap_states = get_reachable_nodes(graph, self.final_states, reversed=True)

        new_states = live_states & non_trap_states
        new_states.add(self.initial_state)

        if minify:
            # No need to alter transitions here, since unused entries in
            # that dict are removed automatically by the minify call
            return self.__class__._minify(
                reachable_states=new_states,
                input_symbols=self.input_symbols,
                transitions=self.transitions,
                initial_state=self.initial_state,
                reachable_final_states=self.final_states & new_states,
                retain_names=retain_names,
            )

        return self.__class__(
            states=new_states,
            input_symbols=self.input_symbols,
            transitions={
                src_state: {
                    symbol: tgt_state
                    for symbol, tgt_state in lookup.items()
                    if tgt_state in non_trap_states
                }
                for src_state, lookup in self.transitions.items()
                if src_state in new_states
            },
            initial_state=self.initial_state,
            final_states=self.final_states & new_states,
            allow_partial=True,
        )

    def _get_trap_state_id(self) -> DFAStateT:
        return next(x for x in count(-1, -1) if x not in self.states)

    def to_complete(self, trap_state: Optional[DFAStateT] = None) -> Self:
        """
        Creates an equivalent complete DFA with trap_state used as the name
        for an added trap state. If trap_state is not passed in, defaults to
        the largest negative integer which is not already a state name.
        If the DFA is already complete, just returns a copy.

        Parameters
        ----------
        trap_state : Optional[DFAStateT], default: None
            Name for custom trap state to be used.

        Returns
        -------
        Self
            An equivalent complete DFA.

        """
        is_partial = any(
            len(lookup) != len(self.input_symbols)
            for lookup in self.transitions.values()
        )

        if not is_partial:
            return self.copy()

        if trap_state is None:
            trap_state = self._get_trap_state_id()
        elif trap_state in self.states:
            raise exceptions.InvalidStateError(
                f"state {trap_state} is already in the state set and "
                "cannot be used as a label for the trap state."
            )

        return self._to_complete(
            input_symbols=self.input_symbols,
            transitions=self.transitions,
            initial_state=self.initial_state,
            final_states=self.final_states,
            trap_state=trap_state,
        )

    @classmethod
    def _to_complete(
        cls: Type[Self],
        *,
        input_symbols: AbstractSet[str],
        transitions: DFATransitionsT,
        initial_state: DFAStateT,
        final_states: AbstractSet[DFAStateT],
        trap_state: DFAStateT,
    ) -> Self:
        """
        Internal helper function taking in a description of a partial DFA and
        returning a corresponding complete DFA.

        Note: This will not work with the description of a complete DFA,
        so be sure the input describes a partial one.
        """

        default_to_trap = {symbol: trap_state for symbol in input_symbols}
        transitions = {
            state: {**default_to_trap, **lookup}
            for state, lookup in transitions.items()
        }
        transitions[trap_state] = default_to_trap

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state,
            final_states=final_states,
            allow_partial=False,
        )

    def _validate_transition_missing_symbols(
        self, start_state: DFAStateT, paths: DFAPathT
    ) -> None:
        """Raise an error if the transition input_symbols are missing."""
        if self.allow_partial:
            return
        for input_symbol in self.input_symbols:
            if input_symbol not in paths:
                raise exceptions.MissingSymbolError(
                    f"state {start_state} is missing transitions "
                    f'for symbol "{input_symbol}"'
                )

    def _validate_transition_invalid_symbols(
        self, start_state: DFAStateT, paths: DFAPathT
    ) -> None:
        """Raise an error if transition input symbols are invalid."""
        for input_symbol in paths.keys():
            if input_symbol not in self.input_symbols:
                raise exceptions.InvalidSymbolError(
                    f"state {start_state} has invalid transition symbol"
                    f' "{input_symbol}"'
                )

    def _validate_transition_start_states(self) -> None:
        """Raise an error if transition start states are missing."""
        for state in self.states:
            if state not in self.transitions:
                raise exceptions.MissingStateError(
                    f"transition start state {state} is missing"
                )

    def _validate_transition_end_states(
        self, start_state: DFAStateT, paths: DFAPathT
    ) -> None:
        """Raise an error if transition end states are invalid."""
        for end_state in paths.values():
            if end_state not in self.states:
                raise exceptions.InvalidStateError(
                    f"end state {end_state} for transition on "
                    f"{start_state} is not valid"
                )

    def _validate_transitions(self, start_state: DFAStateT, paths: DFAPathT) -> None:
        """Raise an error if transitions are missing or invalid."""
        self._validate_transition_missing_symbols(start_state, paths)
        self._validate_transition_invalid_symbols(start_state, paths)
        self._validate_transition_end_states(start_state, paths)

    def validate(self) -> None:
        """
        Raises an exception if this automaton is not internally consistent.

        Raises
        ------
        InvalidStateError
            If this DFA has invalid states in the transition dictionary.
        MissingStateError
            If this DFA has states missing from the transition dictionary.
        InvalidSymbolError
            If this DFA has invalid symbols in the transition dictionary.
        MissingSymbolError
            If this DFA is missing transitions on certain symbols.
        """

        self._validate_transition_start_states()
        for start_state, paths in self.transitions.items():
            self._validate_transitions(start_state, paths)
        self._validate_initial_state()
        self._validate_final_states()

    def _get_next_current_state(
        self, current_state: DFAStateT, input_symbol: str
    ) -> Optional[DFAStateT]:
        """
        Follow the transition for the given input symbol on the current state.

        Return None if transition does not exist.
        """
        if (
            current_state is not None
            and input_symbol in self.transitions[current_state]
        ):
            return self.transitions[current_state][input_symbol]
        return None

    def _check_for_input_rejection(self, current_state: DFAStateT) -> None:
        """Raise an error if the given config indicates rejected input."""
        if current_state not in self.final_states:
            raise exceptions.RejectionException(
                f"the DFA stopped on a non-final state ({current_state})"
            )

    def read_input_stepwise(
        self, input_str: str, ignore_rejection: bool = False
    ) -> Generator[DFAStateT, None, None]:
        """
        Return a generator that yields each step while reading input.

        Parameters
        ----------
        input_str : str
            The input string to read.
        ignore_rejection : bool, default: False
            Whether to throw an exception if the input string is rejected.

        Yields
        ------
        Generator[DFAStateT, None, None]
            A generator that yields the current configuration of the DFA
            after each step of reading input.

        Raises
        ------
        RejectionException
            Raised if this DFA does not accept the input string.
        """
        current_state = self.initial_state

        yield current_state
        for input_symbol in input_str:
            current_state = self._get_next_current_state(current_state, input_symbol)
            yield current_state

        if not ignore_rejection:
            self._check_for_input_rejection(current_state)

    @cached_method
    def _get_digraph(self) -> nx.DiGraph:
        """Return a digraph corresponding to this DFA with transition symbols ignored"""

        graph = nx.DiGraph()
        graph.add_nodes_from(self.states)
        graph.add_edges_from(
            (start_state, end_state)
            for start_state, transition in self.transitions.items()
            for end_state in transition.values()
        )

        return graph

    def minify(self, retain_names: bool = False) -> Self:
        """
        Create a minimal DFA which accepts the same inputs as this DFA.

        First, non-reachable states are removed.
        Then, indistinguishable states are merged using Hopcroft's Algorithm.

        Parameters
        ----------
        retain_names : bool, default: False
            Whether to retain original names when merging states.
            New names are from 0 to n-1.

        Returns
        ------
        Self
            A state-minimal equivalent DFA. May be complete in some cases
            if the input is partial.
        """

        if self.allow_partial:
            # In the case of a partial DFA, we want to try to condense
            # possible trap states before the main minify operation.
            graph = self._get_digraph()
            live_states = get_reachable_nodes(graph, [self.initial_state])
            non_trap_states = get_reachable_nodes(
                graph, self.final_states, reversed=True
            )

            reachable_states = live_states & non_trap_states
            reachable_states.add(self.initial_state)
        else:
            # Compute reachable states and final states
            bfs_states = self.__class__._bfs_states(
                self.initial_state, lambda state: iter(self.transitions[state].items())
            )
            reachable_states = set(bfs_states)

        reachable_final_states = self.final_states & reachable_states

        return self.__class__._minify(
            reachable_states=reachable_states,
            input_symbols=self.input_symbols,
            transitions=self.transitions,
            initial_state=self.initial_state,
            reachable_final_states=reachable_final_states,
            retain_names=retain_names,
        )

    @classmethod
    def _minify(
        cls: Type[Self],
        *,
        reachable_states: AbstractSet[DFAStateT],
        input_symbols: AbstractSet[str],
        transitions: DFATransitionsT,
        initial_state: DFAStateT,
        reachable_final_states: AbstractSet[DFAStateT],
        retain_names: bool,
    ) -> Self:
        """
        Minify helper function. DFA data passed in must have no unreachable states.
        If the input DFA is partial, then the result is also a partial DFA
        """

        reachable_states = set(reachable_states)

        # Per input-symbol backmap (tgt -> origin states)
        transition_back_map: Dict[str, Dict[DFAStateT, List[DFAStateT]]] = {
            symbol: {end_state: [] for end_state in reachable_states}
            for symbol in input_symbols
        }

        trap_state = None

        for start_state, path in transitions.items():
            if start_state in reachable_states:
                for symbol in input_symbols:
                    end_state = path.get(symbol)
                    if end_state is not None:
                        symbol_dict = transition_back_map[symbol]
                        # If statement here needed to ignore certain transitions
                        # for non-reachable states
                        if end_state in symbol_dict:
                            symbol_dict[end_state].append(start_state)
                    else:
                        # Add trap state if needed
                        if trap_state is None:
                            trap_state = next(
                                x for x in count(-1, -1) if x not in reachable_states
                            )
                            for trap_symbol in input_symbols:
                                transition_back_map[trap_symbol][trap_state] = [
                                    trap_state
                                ]

                            reachable_states.add(trap_state)

                        transition_back_map[symbol][trap_state].append(start_state)

        # Set up equivalence class data structure
        eq_classes = PartitionRefinement(reachable_states)
        refinement = eq_classes.refine(reachable_final_states)

        final_states_id = (
            refinement[0][0] if refinement else next(iter(eq_classes.get_set_ids()))
        )

        origin_dicts = tuple(transition_back_map.values())
        processing = {final_states_id}

        while processing:
            # Save a copy of the set, since it could get modified while executing
            active_state = tuple(eq_classes.get_set_by_id(processing.pop()))
            for origin_dict in origin_dicts:
                states_that_move_into_active_state = chain.from_iterable(
                    origin_dict[end_state] for end_state in active_state
                )

                # Refine set partition by states moving into current active one
                new_eq_class_pairs = eq_classes.refine(
                    states_that_move_into_active_state
                )

                for YintX_id, YdiffX_id in new_eq_class_pairs:
                    # Only adding one id to processing, since the other is already there
                    if YdiffX_id in processing:
                        processing.add(YintX_id)
                    else:
                        if len(eq_classes.get_set_by_id(YintX_id)) <= len(
                            eq_classes.get_set_by_id(YdiffX_id)
                        ):
                            processing.add(YintX_id)
                        else:
                            processing.add(YdiffX_id)

        # now eq_classes are good to go, make them a list for ordering
        eq_class_name_pairs: List[Tuple[DFAStateT, Set[DFAStateT]]] = (
            [(frozenset(eq), eq) for eq in eq_classes.get_sets()]
            if retain_names
            else list(enumerate(eq_classes.get_sets()))
        )

        # need a backmap to prevent constant calls to index
        back_map = {
            state: name
            for name, eq in eq_class_name_pairs
            for state in eq
            if trap_state not in eq
        }

        # If only one equivalence class with the trap state,
        # return empty language.
        if not back_map:
            return cls.empty_language(input_symbols)

        new_input_symbols = input_symbols
        new_states = frozenset(back_map.values())
        new_initial_state = back_map[initial_state]
        new_final_states = frozenset(back_map[acc] for acc in reachable_final_states)
        new_transitions = {}

        for name, eq in eq_class_name_pairs:
            # For trap state, can just leave out
            if trap_state in eq:
                continue

            eq_class_rep = next(iter(eq))

            inner_transition_dict_old = transitions[eq_class_rep]
            new_transitions[name] = {
                letter: back_map[inner_transition_dict_old[letter]]
                for letter in inner_transition_dict_old.keys()
                if inner_transition_dict_old[letter] in back_map.keys()
            }

        allow_partial = any(
            len(lookup) != len(input_symbols) for lookup in new_transitions.values()
        )
        return cls(
            states=new_states,
            input_symbols=new_input_symbols,
            transitions=new_transitions,
            initial_state=new_initial_state,
            final_states=new_final_states,
            allow_partial=allow_partial,
        )

    def union(
        self, other: DFA, *, retain_names: bool = False, minify: bool = True
    ) -> Self:
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the union of L1 and L2.

        Minifies by default. Unreachable states are always removed.
        If either input DFA is partial, the result is partial.

        Parameters
        ----------
        other : DFA
            The DFA we want to take a union with.
        retain_names : bool, default: False
            Whether to retain state names through the union and optional minify.
        minify : bool, default: True
            Whether to minify the result of the union of the two DFAs.

        Returns
        ------
        Self
            A DFA accepting the union of the two input DFAs. State minimal by
            default.
        """

        def union_function(state_pair: Tuple[DFAStateT, DFAStateT]) -> bool:
            q_a, q_b = state_pair
            return q_a in self.final_states or q_b in other.final_states

        initial_state, expand_state_fn = self.__class__._cross_product(
            self, other, lhs_relevant=True, rhs_relevant=True
        )

        return self.__class__._expand_dfa(
            union_function,
            initial_state,
            expand_state_fn,
            self.input_symbols,
            retain_names=retain_names,
            minify=minify,
        )

    def intersection(
        self, other: DFA, *, retain_names: bool = False, minify: bool = True
    ) -> Self:
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the intersection of L1 and L2.

        Minifies by default. Unreachable states are always removed.
        If either input DFA is partial, the result is partial.

        Parameters
        ----------
        other : DFA
            The DFA we want to take a intersection with.
        retain_names : bool, default: False
            Whether to retain state names through the intersection and optional minify.
        minify : bool, default: True
            Whether to minify the result of the intersection of the two DFAs.

        Returns
        ------
        Self
            A DFA accepting the intersection of the two input DFAs. State minimal by
            default.
        """

        def intersection_function(state_pair: Tuple[DFAStateT, DFAStateT]) -> bool:
            q_a, q_b = state_pair
            return q_a in self.final_states and q_b in other.final_states

        initial_state, expand_state_fn = self.__class__._cross_product(
            self, other, lhs_relevant=False, rhs_relevant=False
        )

        return self.__class__._expand_dfa(
            intersection_function,
            initial_state,
            expand_state_fn,
            self.input_symbols,
            retain_names=retain_names,
            minify=minify,
        )

    def difference(
        self, other: DFA, *, retain_names: bool = False, minify: bool = True
    ) -> Self:
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the difference of L1 and L2.

        Minifies by default. Unreachable states are always removed.
        If either input DFA is partial, the result is partial.

        Parameters
        ----------
        other : DFA
            The DFA we want to take a difference with.
        retain_names : bool, default: False
            Whether to retain state names through the difference and optional minify.
        minify : bool, default: True
            Whether to minify the result of the difference of the two DFAs.

        Returns
        ------
        Self
            A DFA accepting the difference of the two input DFAs. State minimal by
            default.
        """

        def difference_function(state_pair: Tuple[DFAStateT, DFAStateT]) -> bool:
            q_a, q_b = state_pair
            return q_a in self.final_states and q_b not in other.final_states

        initial_state, expand_state_fn = self.__class__._cross_product(
            self, other, lhs_relevant=False, rhs_relevant=True
        )

        return self.__class__._expand_dfa(
            difference_function,
            initial_state,
            expand_state_fn,
            self.input_symbols,
            retain_names=retain_names,
            minify=minify,
        )

    def symmetric_difference(
        self, other: DFA, *, retain_names: bool = False, minify: bool = True
    ) -> Self:
        """
        Takes as input two DFAs M1 and M2 which
        accept languages L1 and L2 respectively.
        Returns a DFA which accepts the symmetric difference of L1 and L2.

        Minifies by default. Unreachable states are always removed.
        If either input DFA is partial, the result is partial.

        Parameters
        ----------
        other : DFA
            The DFA we want to take a symmetric difference with.
        retain_names : bool, default: False
            Whether to retain state names through the symmetric difference and optional
            minify.
        minify : bool, default: True
            Whether to minify the result of the symmetric difference of the two DFAs.

        Returns
        ------
        Self
            A DFA accepting the symmetric difference of the two input DFAs. State
            minimal by default.
        """

        def symmetric_difference_function(
            state_pair: Tuple[DFAStateT, DFAStateT],
        ) -> bool:
            q_a, q_b = state_pair
            return (q_a in self.final_states) ^ (q_b in other.final_states)

        initial_state, expand_state_fn = self.__class__._cross_product(
            self, other, lhs_relevant=True, rhs_relevant=True
        )

        return self.__class__._expand_dfa(
            symmetric_difference_function,
            initial_state,
            expand_state_fn,
            self.input_symbols,
            retain_names=retain_names,
            minify=minify,
        )

    def complement(self, *, retain_names: bool = False, minify: bool = True) -> Self:
        """
        Creates a DFA which accepts an input if and only if the old one does not.
        Minifies by default. Unreachable states are always removed. Partial DFAs
        are converted into complete ones.

        Parameters
        ----------
        retain_names : bool, default: False
            Whether to retain state names through the complement and optional
            minify.
        minify : bool, default: True
            Whether to minify the result of the complement of the input DFA.

        Returns
        ------
        Self
            A DFA accepting the complement of the input DFA. State
            minimal by default.
        """

        # We can't do much here, we must turn it into a complete DFA
        complete_dfa: Self = self.to_complete() if self.allow_partial else self

        if minify:
            bfs_states = self.__class__._bfs_states(
                complete_dfa.initial_state,
                lambda state: iter(complete_dfa.transitions[state].items()),
            )

            reachable_states = set(bfs_states)
            reachable_final_states = complete_dfa.final_states & reachable_states

            return complete_dfa.__class__._minify(
                reachable_states=reachable_states,
                input_symbols=complete_dfa.input_symbols,
                transitions=complete_dfa.transitions,
                initial_state=complete_dfa.initial_state,
                reachable_final_states=reachable_states - reachable_final_states,
                retain_names=retain_names,
            )

        return complete_dfa.__class__(
            states=complete_dfa.states,
            input_symbols=complete_dfa.input_symbols,
            transitions=complete_dfa.transitions,
            initial_state=complete_dfa.initial_state,
            final_states=complete_dfa.states - complete_dfa.final_states,
        )

    @staticmethod
    def _bfs_edges(
        initial_state: DFAStateT,
        expand_state_fn: ExpandStateFn,
    ) -> Generator[Tuple[DFAStateT, DFASymbolT, DFAStateT], None, None]:
        """
        Emits the edges (src_state, label, tgt_state) visited by BFS from the
        initial_state. Computes subsequent states using the function expand_state_fn.
        """
        visited_set = {initial_state}
        queue: Deque[Tuple[DFAStateT, DFAStateT]] = deque([initial_state])

        while queue:
            curr_state = queue.popleft()

            for chr, tgt_state in expand_state_fn(curr_state):
                yield curr_state, chr, tgt_state

                if tgt_state not in visited_set:
                    visited_set.add(tgt_state)
                    queue.append(tgt_state)

    @staticmethod
    def _bfs_states(
        initial_state: DFAStateT, expand_state_fn: ExpandStateFn
    ) -> Generator[DFAStateT, None, None]:
        """
        Emits the states visited by BFS from the initial_state.
        Computes subsequent states using the function expand_state_fn.
        """
        yield initial_state
        visited_set = {initial_state}
        queue = deque([initial_state])

        while queue:
            curr_state = queue.popleft()

            for _, tgt_state in expand_state_fn(curr_state):
                if tgt_state not in visited_set:
                    yield tgt_state
                    visited_set.add(tgt_state)
                    queue.append(tgt_state)

    @classmethod
    def _expand_dfa(
        cls: Type[Self],
        final_state_fn: IsFinalStateFn,
        initial_state: DFAStateT,
        expand_state_fn: ExpandStateFn,
        input_symbols: AbstractSet[DFASymbolT],
        retain_names: bool = False,
        minify: bool = True,
    ) -> Self:
        """
        Constructs the DFA by expanding from the initial_state using the expand_state_fn
        function. The function final_state_fn must return True for the final states.
        The function expand_state_fn must return an iterator with the successors
        of each state in the product.

        If minify is set to True, the returned DFA is minified. If retain_names
        is set to False, states are renamed.
        """

        if retain_names:

            def get_name_original(state: DFAStateT) -> DFAStateT:
                return state

            get_name: Callable[[DFAStateT], DFAStateT] = get_name_original
        else:
            get_name = get_renaming_function(count(0))

        initial_state_name = get_name(initial_state)

        transitions: Dict[DFAStateT, Dict[str, DFAStateT]] = {initial_state_name: {}}
        states = {initial_state_name}
        final_states = {initial_state_name} if final_state_fn(initial_state) else set()

        for cur_state, chr, tgt_state in cls._bfs_edges(initial_state, expand_state_fn):
            cur_state_name = get_name(cur_state)
            tgt_state_name = get_name(tgt_state)

            if tgt_state_name not in states:
                states.add(tgt_state_name)
                transitions[tgt_state_name] = {}

            transitions[cur_state_name][chr] = tgt_state_name

            if final_state_fn(tgt_state):
                final_states.add(tgt_state_name)

        if minify:
            # From the construction, the states/final states are reachable
            return cls._minify(
                reachable_states=states,
                input_symbols=input_symbols,
                transitions=transitions,
                initial_state=initial_state_name,
                reachable_final_states=final_states,
                retain_names=retain_names,
            )

        # This method is generic and is used to implement (mostly) binary
        #   operators; we don't offer the option to return complete/partial
        #   DFAs; instead this is inferred directly from the structure of the
        #   resulting DFA. But for these operations, the result will be a
        #   complete DFA if the inputs are also complete DFAs
        is_partial = any(
            len(lookup) != len(input_symbols) for lookup in transitions.values()
        )
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state_name,
            final_states=final_states,
            allow_partial=is_partial,
        )

    @classmethod
    def _find_state(
        cls: Type[Self],
        target_state_fn: TargetStateFn,
        initial_state: DFAStateT,
        expand_state_fn: ExpandStateFn,
    ) -> bool:
        """
        Searches for a target state in the DFA without explicitly constructing
        it. The function target_state_fn should return True for states that are being
        searched for. The expand_state_fn should return an iterator with the
        successors of each state in the product.
        """
        bfs_states = cls._bfs_states(initial_state, expand_state_fn)
        return any(target_state_fn(state) for state in bfs_states)

    @staticmethod
    def _cross_product(
        lhs: DFA, rhs: DFA, *, lhs_relevant: bool, rhs_relevant: bool
    ) -> Tuple[DFAStateT, ExpandStateFn]:
        """
        Builds the cross product between the two DFAs.
        Keeps expanding the lhs and rhs if the trap state is encountered based on the
        lhs_relevant and rhs_relevant booleans.
        """

        if lhs.input_symbols != rhs.input_symbols:
            raise exceptions.SymbolMismatchError(
                "The input symbols between the two given DFAs do not match"
            )

        trap_a = lhs._get_trap_state_id()
        trap_b = rhs._get_trap_state_id()

        # TODO if anyone ever has issues with this function being too slow,
        # there is a way to speed this up in special cases with some complicated
        # logic with the lhs_relevant and rhs_relevant booleans and the inner
        # chr loop below.

        def expand_state_fn(state: DFAStateT) -> ExpandStateReturnType:
            q_a, q_b = state
            # Have to use .get() here since these might be trap states
            transitions_a = lhs.transitions.get(q_a, {})
            transitions_b = rhs.transitions.get(q_b, {})

            for chr in transitions_a.keys() | transitions_b.keys():
                # Equivalent to reaching the trap state, skip ahead if irrelevant
                if not lhs_relevant and chr not in transitions_a:
                    continue
                elif not rhs_relevant and chr not in transitions_b:
                    continue

                yield (
                    chr,
                    (
                        transitions_a.get(chr, trap_a),
                        transitions_b.get(chr, trap_b),
                    ),
                )

        initial_state = (lhs.initial_state, rhs.initial_state)

        return initial_state, expand_state_fn

    def issubset(self, other: DFA) -> bool:
        """
        Returns True if the language accepted by self is a subset of that of other.

        Parameters
        ----------
        other : DFA
            The other DFA we are comparing our language against.

        Returns
        ------
        bool
            True if self is a subset of other, False otherwise.
        """

        def subset_state_fn(state_pair: Tuple[DFAStateT, DFAStateT]) -> bool:
            """Check for reachable state that is counterexample to subset"""
            q_a, q_b = state_pair
            return q_a in self.final_states and q_b not in other.final_states

        initial_state, expand_state_fn = self.__class__._cross_product(
            self, other, lhs_relevant=False, rhs_relevant=True
        )

        return not self.__class__._find_state(
            subset_state_fn, initial_state, expand_state_fn
        )

    def issuperset(self, other: DFA) -> bool:
        """
        Returns True if the language accepted by self is a superset of that of other.

        Parameters
        ----------
        other : DFA
            The other DFA we are comparing our language against.

        Returns
        ------
        bool
            True if self is a superset of other, False otherwise.
        """
        return other.issubset(self)

    def isdisjoint(self, other: DFA) -> bool:
        """
        Returns True if the language accepted by self is disjoint from that of other.

        Parameters
        ----------
        other : DFA
            The other DFA we are comparing our language against.

        Returns
        ------
        bool
            True if self is disjoint from other, False otherwise.
        """

        def disjoint_state_fn(state_pair: Tuple[DFAStateT, DFAStateT]) -> bool:
            """Check for reachable state that is counterexample to disjointness"""
            q_a, q_b = state_pair
            return q_a in self.final_states and q_b in other.final_states

        initial_state, expand_state_fn = self.__class__._cross_product(
            self, other, lhs_relevant=False, rhs_relevant=False
        )

        return not self.__class__._find_state(
            disjoint_state_fn,
            initial_state,
            expand_state_fn,
        )

    @cached_method
    def isempty(self) -> bool:
        """
        Returns True if the language accepted by self is empty.

        Returns
        ------
        bool
            True if self accepts the empty language, False otherwise.
        """
        return not self.__class__._find_state(
            self.final_states.__contains__,
            self.initial_state,
            lambda state: iter(self.transitions[state].items()),
        )

    @cached_method
    def isfinite(self) -> bool:
        """
        Returns True if the language accepted by self is finite.

        Returns
        ------
        bool
            True if self accepts a finite language, False otherwise.
        """
        try:
            return self.maximum_word_length() is not None
        except exceptions.EmptyLanguageException:
            return True

    def random_word(self, k: int, *, seed: Optional[int] = None) -> str:
        """
        Returns a random word of length k accepted by self.

        Parameters
        ----------
        k : int
            The length of the desired word.
        seed : Optional[int], default: None
            The random seed to use for the sampling of the random word.

        Returns
        ------
        str
            A uniformly random word of length k accepted by the DFA self.

        Raises
        ------
        ValueError
            If this DFA does not accept any words of length k.
        """
        self._populate_count_cache_up_to_len(k)
        state = self.initial_state
        if self._count_cache[k][state] == 0:
            raise ValueError(f"Language has no words of length {k}")

        result = []
        rng = Random(seed)
        for remaining in range(k, 0, -1):
            total = self._count_cache[remaining][state]
            choice = rng.randint(0, total - 1)
            transition = self.transitions[state]
            for symbol, next_state in transition.items():
                next_state_count = self._count_cache[remaining - 1][next_state]
                if choice < next_state_count:
                    result.append(symbol)
                    state = next_state
                    break
                choice -= next_state_count

        assert state in self.final_states
        return "".join(result)

    def predecessor(
        self,
        input_str: str,
        *,
        strict: bool = True,
        key: Optional[Callable[[Any], Any]] = None,
        min_length: int = 0,
        max_length: Optional[int] = None,
    ) -> Optional[str]:
        """
        Returns the first string accepted by the DFA that comes before
        the input string in lexicographical order.

        Parameters
        ----------
        input_str : str
            The starting input string.
        strict : bool, default: True
            If set to false and input_str is accepted by the DFA, input_str will be
            returned.
        key : Optional[Callable], default: None
            Function for defining custom lexicographical ordering. Defaults to using
            the standard string ordering.
        min_length : int, default: 0
            Limits generation to words with at least the given length.
        max_length : Optional[int], default: None
            Limits generation to words with at most the given length.

        Returns
        ------
        str
            The first string accepted by the DFA lexicographically before input_string.

        Raises
        ------
        InfiniteLanguageException
            Raised if the language accepted by self is infinite, as we cannot
            generate predecessors in this case.
        """
        for word in self.predecessors(
            input_str,
            strict=strict,
            key=key,
            min_length=min_length,
            max_length=max_length,
        ):
            return word
        return None

    def predecessors(
        self,
        input_str: str,
        *,
        strict: bool = True,
        key: Optional[Callable[[Any], Any]] = None,
        min_length: int = 0,
        max_length: Optional[int] = None,
    ) -> Generator[str, None, None]:
        """
        Generates all strings that come before the input string
        in lexicographical order.

        Parameters
        ----------
        input_str : str
            The starting input string.
        strict : bool, default: True
            If set to false and input_str is accepted by the DFA, input_str will be
            returned.
        key : Optional[Callable], default: None
            Function for defining custom lexicographical ordering. Defaults to using
            the standard string ordering.
        min_length : int, default: 0
            Limits generation to words with at least the given length.
        max_length : Optional[int], default: None
            Limits generation to words with at most the given length.

        Returns
        ------
        Generator[str, None, None]
            A generator for all strings that come before the input string in
            lexicographical order.

        Raises
        ------
        InfiniteLanguageException
            Raised if the language accepted by self is infinite, as we cannot
            generate predecessors in this case.
        """
        yield from self.successors(
            input_str,
            strict=strict,
            reverse=True,
            key=key,
            min_length=min_length,
            max_length=max_length,
        )

    def successor(
        self,
        input_str: Optional[str],
        *,
        strict: bool = True,
        key: Optional[Callable[[Any], Any]] = None,
        min_length: int = 0,
        max_length: Optional[int] = None,
    ) -> Optional[str]:
        """
        Returns the first string accepted by the DFA that comes after
        the input string in lexicographical order.

        Parameters
        ----------
        input_str : Optional[str]
            The starting input string. If None, will generate all words.
        strict : bool, default: True
            If set to false and input_str is accepted by the DFA, input_str will be
            returned.
        key : Optional[Callable], default: None
            Function for defining custom lexicographical ordering. Defaults to using
            the standard string ordering.
        min_length : int, default: 0
            Limits generation to words with at least the given length.
        max_length : Optional[int], default: None
            Limits generation to words with at most the given length.

        Returns
        ------
        str
            The first string accepted by the DFA lexicographically before input_string.
        """
        for word in self.successors(
            input_str,
            strict=strict,
            key=key,
            min_length=min_length,
            max_length=max_length,
        ):
            return word
        return None

    def successors(
        self,
        input_str: Optional[str],
        *,
        strict: bool = True,
        key: Optional[Callable[[Any], Any]] = None,
        reverse: bool = False,
        min_length: int = 0,
        max_length: Optional[int] = None,
    ) -> Generator[str, None, None]:
        """
        Generates all strings that come after the input string
        in lexicographical order.

        Parameters
        ----------
        input_str : Optional[str]
            The starting input string. If None, will generate all words.
        strict : bool, default: True
            If set to false and input_str is accepted by the DFA, input_str will be
            returned.
        key : Optional[Callable], default: None
            Function for defining custom lexicographical ordering. Defaults to using
            the standard string ordering.
        reverse : bool, default: False
            If True, then predecessors will be generated instead of successors.
        min_length : int, default: 0
            Limits generation to words with at least the given length.
        max_length : Optional[int], default: None
            Limits generation to words with at most the given length.

        Returns
        ------
        Generator[str, None, None]
            A generator for all strings that come after the input string in
            lexicographical order.

        """
        # A predecessor for a finite string may be infinite but a successor for
        # a finite string is always finite
        if reverse and not self.isfinite():
            raise exceptions.InfiniteLanguageException(
                "Predecessors cannot be computed for infinite languages"
            )
        graph = self._get_digraph()
        coaccessible_nodes = get_reachable_nodes(
            graph, self.final_states, reversed=True
        )

        # Precomputations and setup
        include_input = not strict
        sorted_symbols = sorted(self.input_symbols, reverse=reverse, key=key)
        symbol_succ: Dict[str, Optional[str]] = {
            symbol_a: symbol_b for symbol_a, symbol_b in pairwise(sorted_symbols)
        }
        symbol_succ[sorted_symbols[-1]] = None
        # Special case for None
        state_stack: Deque[Optional[DFAStateT]] = deque(
            [self.initial_state]
            if input_str is None
            else self.read_input_stepwise(input_str, ignore_rejection=True)
        )
        char_stack: Deque[str] = deque("" if input_str is None else input_str)
        first_symbol = sorted_symbols[0]
        # For predecessors we need to special case the input string None
        candidate = None if reverse and input_str is not None else first_symbol
        # If input_str is None then we yield on first value found no matter the
        # value of include_input
        should_yield = include_input or input_str is None

        # Iterative preorder/postorder (depends on reverse) traversal that
        # yields on final states
        while char_stack or candidate is not None:
            state = state_stack[-1]
            # Successors yield here
            if (
                not reverse
                and should_yield
                and min_length <= len(char_stack)
                and (max_length is None or len(char_stack) <= max_length)
                and candidate == first_symbol
                and state in self.final_states
            ):
                yield "".join(char_stack)
            candidate_state = (
                None
                if candidate is None
                else self._get_next_current_state(state, candidate)
            )
            # Traverse to child if candidate is viable
            if candidate_state in coaccessible_nodes and (
                max_length is None or len(char_stack) < max_length
            ):
                state_stack.append(candidate_state)
                char_stack.append(cast(str, candidate))
                candidate = first_symbol
            else:
                # Predecessors yield here
                if (
                    reverse
                    and should_yield
                    and min_length <= len(char_stack)
                    and (max_length is None or len(char_stack) <= max_length)
                    and candidate is None
                    and state in self.final_states
                ):
                    yield "".join(char_stack)
                # Candidate is None means no more children to explore, so
                # traverse to parent
                if candidate is None:
                    state = state_stack.pop()
                    candidate = char_stack.pop()
                candidate = symbol_succ[candidate]
            should_yield = True
        # Predecessor yields here for empty string
        if (
            reverse
            and should_yield
            and min_length <= len(char_stack)
            and (max_length is None or len(char_stack) <= max_length)
            and candidate is None
            and state in self.final_states
        ):
            yield "".join(char_stack)

    def count_words_of_length(self, k: int) -> int:
        """
        Returns count of words of length k accepted by the DFA.

        Parameters
        ----------
        k : int
            The desired word length.

        Returns
        ------
        int
            The number of words of length k accepted by self.
        """
        self._populate_count_cache_up_to_len(k)
        return self._count_cache[k][self.initial_state]

    def _populate_count_cache_up_to_len(self, k: int) -> None:
        """
        Populate count cache up to length k
        """

        while len(self._count_cache) <= k:
            i = len(self._count_cache)
            self._count_cache.append(defaultdict(int))
            level = self._count_cache[i]
            if i == 0:
                level.update({state: 1 for state in self.final_states})
            else:
                prev_level = self._count_cache[i - 1]
                level.update(
                    {
                        state: sum(
                            prev_level[suffix_state]
                            for suffix_state in self.transitions[state].values()
                        )
                        for state in self.states
                    }
                )

    def words_of_length(self, k: int) -> Generator[str, None, None]:
        """
        Generates all words of length `k` in the language accepted by the DFA.

        Parameters
        ----------
        k : int
            The desired word length.

        Returns
        ------
        Generator[str, None, None]
            A generator for all words of length k accepted by the DFA.
        """
        self._populate_word_cache_up_to_len(k)
        for word in self._word_cache[k][self.initial_state]:
            yield word

    def _populate_word_cache_up_to_len(self, k: int) -> None:
        """
        Populate word cache up to length k
        """

        # Weird construction to account for partial DFAs.
        sorted_transition_symbols = {
            state: sorted(lookup.keys()) for state, lookup in self.transitions.items()
        }

        while len(self._word_cache) <= k:
            i = len(self._word_cache)
            self._word_cache.append(defaultdict(list))
            level = self._word_cache[i]
            if i == 0:
                level.update({state: [""] for state in self.final_states})
            else:
                prev_level = self._word_cache[i - 1]
                level.update(
                    {
                        state: [
                            symbol + word
                            for symbol in sorted_transition_symbols[state]
                            for word in prev_level[self.transitions[state][symbol]]
                        ]
                        for state in self.states
                    }
                )

    @cached_method
    def cardinality(self) -> int:
        """
        Returns the cardinality of the language represented by the DFA.

        Returns
        ------
        int
            The cardinality of the language accepted by self.

        Raises
        ------
        InfiniteLanguageException
            Raised if self accepts an infinite language.
        """
        try:
            i = self.minimum_word_length()
        except exceptions.EmptyLanguageException:
            return 0
        limit = self.maximum_word_length()
        if limit is None:
            raise exceptions.InfiniteLanguageException(
                "The language represented by the DFA is infinite."
            )
        return sum(self.count_words_of_length(j) for j in range(i, limit + 1))

    @cached_method
    def minimum_word_length(self) -> int:
        """
        Returns the length of the shortest word in the language accepted by the DFA.

        Returns
        ------
        int
            The length of the shortest word accepted by self.

        Raises
        ------
        EmptyLanguageException
            Raised if self accepts an empty language.
        """
        queue: Deque[DFAStateT] = deque()
        distances: Dict[DFAStateT, int] = {self.initial_state: 0}
        queue.append(self.initial_state)
        while queue:
            state = queue.popleft()
            if state in self.final_states:
                return distances[state]
            for next_state in self.transitions[state].values():
                if next_state not in distances:
                    distances[next_state] = distances[state] + 1
                    queue.append(next_state)
        raise exceptions.EmptyLanguageException(
            "The language represented by the DFA is empty"
        )

    @cached_method
    def maximum_word_length(self) -> Optional[int]:
        """
        Returns the length of the longest word in the language accepted by the DFA
        In the case of infinite languages, `None` is returned.

        Returns
        ------
        Optional[int]
            The length of the longest word accepted by self. None if the language
            is infinite.

        Raises
        ------
        EmptyLanguageException
            Raised if self accepts the empty language.
        """
        if self.isempty():
            raise exceptions.EmptyLanguageException(
                "The language represented by the DFA is empty"
            )
        graph = self._get_digraph()

        accessible_nodes = get_reachable_nodes(graph, [self.initial_state])
        coaccessible_nodes = get_reachable_nodes(
            graph, self.final_states, reversed=True
        )

        important_nodes = accessible_nodes.intersection(coaccessible_nodes)
        subgraph = graph.subgraph(important_nodes)
        try:
            return nx.dag_longest_path_length(subgraph)
        except nx.exception.NetworkXUnfeasible:
            return None

    @classmethod
    def from_prefix(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        prefix: str,
        *,
        contains: bool = True,
        as_partial: bool = True,
    ) -> Self:
        """
        Directly computes the minimal DFA accepting strings with the
        given prefix. If `contains` is set to `False` then the complement is
        constructed instead.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        prefix : str
            The prefix of strings that are accepted by this DFA.
        contains : bool, default: True
            Whether or not to construct the compliment DFA.
        as_partial : bool, default: True
            Whether or not to construct this DFA as a partial DFA.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        last_state = len(prefix)
        transitions = {i: {char: i + 1} for i, char in enumerate(prefix)}
        transitions[last_state] = {symbol: last_state for symbol in input_symbols}

        # Can't construct this as a partial if we need to take the complement
        if not as_partial or not contains:
            err_state = -1
            for state_path in transitions.values():
                for symbol in input_symbols:
                    state_path.setdefault(symbol, err_state)
            transitions[err_state] = {symbol: err_state for symbol in input_symbols}

        states = frozenset(transitions.keys())
        final_states = {last_state}

        is_partial = any(
            len(lookup) != len(input_symbols) for lookup in transitions.values()
        )

        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states,
            allow_partial=is_partial,
        )

    @classmethod
    def from_suffix(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        suffix: str,
        *,
        contains: bool = True,
    ) -> Self:
        """
        Directly computes the minimal DFA recognizing strings with the
        given suffix. If `contains` is set to `False`, then the complement
        is constructed instead.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        suffix : str
            The suffix of strings that are accepted by this DFA.
        contains : bool, default: True
            Whether or not to construct the compliment DFA.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        return cls.from_substring(
            input_symbols, suffix, contains=contains, must_be_suffix=True
        )

    @classmethod
    def from_substring(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        substring: str,
        *,
        contains: bool = True,
        must_be_suffix: bool = False,
    ) -> Self:
        """
        Directly computes the minimal DFA recognizing strings containing the
        given substring. If `contains` is set to `False` then the complement
        is constructed instead. If `must_be_suffix` is set to `True`, then
        the substring must be a suffix instead.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        substring : str
            The substring of strings that are accepted by this DFA.
        contains : bool, default: True
            Whether or to construct the compliment DFA.
        must_be_suffix : bool, default: False
            Whether or not the target substring must be a suffix.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        transitions: Dict[DFAStateT, Dict[str, DFAStateT]] = {
            i: {} for i in range(len(substring))
        }
        transitions[len(substring)] = {
            symbol: len(substring) for symbol in input_symbols
        }

        # Computing failure function for partial matches as is done in the
        # Knuth-Morris-Pratt string algorithm so we can quickly compute the
        # next state from another state
        # This table only holds integers, so array.array is used to save memory.
        kmp_table = array.array("i", (-1 for _ in substring))
        candidate = 0
        for i, char in enumerate(substring):
            if i == 0:
                continue
            elif char == substring[candidate]:
                kmp_table[i] = kmp_table[candidate]
            else:
                kmp_table[i] = candidate
                while candidate >= 0 and char != substring[candidate]:
                    candidate = kmp_table[candidate]
            candidate += 1
        kmp_table.append(candidate)

        limit = len(substring) + 1 if must_be_suffix else len(substring)
        for i in range(limit):
            prefix_dict = transitions.setdefault(i, {})
            for symbol in input_symbols:
                # Look for next state after reading in the given input symbol
                candidate = i if i < len(substring) else kmp_table[i]
                while candidate != -1 and substring[candidate] != symbol:
                    candidate = kmp_table[candidate]
                candidate += 1
                prefix_dict[symbol] = candidate

        states = frozenset(transitions.keys())
        final_states = {len(substring)}
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states,
        )

    @classmethod
    def from_substrings(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        substrings: AbstractSet[str],
        *,
        contains: bool = True,
        must_be_suffix: bool = False,
    ) -> Self:
        """
        Directly computes a DFA recognizing strings containing at least one of the
        given substrings. The implementation is based on the Aho-Corasick
        string-searching algorithm. If `contains` is set to `False`, then the
        complement is constructed instead. If `must_be_suffix` is set to `True`,
        then the each substring must be a suffix instead.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        substrings : str
            The set of strings to be recognized by this DFA.
        contains : bool, default: True
            Whether or to construct the compliment DFA.
        must_be_suffix : bool, default: False
            Whether or not the target substrings must be a suffix.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """

        class OutNode:
            def __init__(self, keyword: str, next_node: Optional[OutNode]) -> None:
                self.keyword: str = keyword
                self.successor: Optional[OutNode] = next_node

        class Node:
            def __init__(self) -> None:
                self.out: Optional[OutNode] = None
                self.fail: Optional[Node] = None
                self.successors: Dict[str, Node] = {}

        root = Node()
        labels = {id(root): 0}
        final_states = set()
        for substring in substrings:
            current_node = root
            for symbol in substring:
                current_node.successors.setdefault(symbol, Node())
                current_node = current_node.successors[symbol]
                labels.setdefault(id(current_node), len(labels))
            current_node.out = OutNode(substring, None)

        queue = deque(root.successors.values())
        while queue:
            current_node = queue.popleft()
            for symbol, successor in current_node.successors.items():
                queue.append(successor)

                st = current_node.fail
                while st is not None and symbol not in st.successors:
                    st = st.fail

                if st is None:
                    st = root

                successor.fail = st.successors.get(symbol, None)

                if successor.fail is not None:
                    if successor.out is None:
                        successor.out = successor.fail.out
                    else:
                        out = successor.out
                        while out.successor is not None:
                            out = out.successor
                        out.successor = successor.fail.out

        transitions: Dict[DFAStateT, Dict[str, DFAStateT]] = {}

        queue = deque([root])
        while queue:
            current_node = queue.popleft()
            state = labels[id(current_node)]
            if current_node.out is not None:
                final_states.add(state)
            current_transitions = {}
            for symbol in input_symbols:
                if symbol in current_node.successors:
                    queue.append(current_node.successors[symbol])
                parent_node: Optional[Node] = current_node
                while parent_node is not None and symbol not in parent_node.successors:
                    parent_node = parent_node.fail
                if parent_node is None:
                    parent_node = root
                successor = parent_node.successors.get(symbol, root)
                current_transitions[symbol] = labels[id(successor)]

            transitions[state] = current_transitions

        if not must_be_suffix:
            end_state = len(transitions)
            transitions[end_state] = {symbol: end_state for symbol in input_symbols}
            for state in final_states:
                transitions[state] = {symbol: end_state for symbol in input_symbols}
            final_states.add(end_state)

        states = frozenset(transitions.keys())
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states,
        )

    @classmethod
    def from_subsequence(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        subsequence: str,
        *,
        contains: bool = True,
    ) -> Self:
        """
        Directly computes the minimal DFA recognizing strings containing the
        given subsequence. If `contains` is set to `False`, then the complement
        is constructed instead.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        subsequence : str
            The target subsequence of strings that are accepted by this DFA.
        contains : bool, default: True
            Whether or to construct the compliment DFA.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        transitions = {0: {symbol: 0 for symbol in input_symbols}}

        for prev_state, char in enumerate(subsequence):
            next_state = prev_state + 1
            transitions[next_state] = {symbol: next_state for symbol in input_symbols}
            transitions[prev_state][char] = next_state

        states = frozenset(transitions.keys())
        final_states = {len(subsequence)}
        return cls(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states if contains else states - final_states,
        )

    @classmethod
    def of_length(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        *,
        min_length: int = 0,
        max_length: Optional[int] = None,
        symbols_to_count: Optional[AbstractSet[str]] = None,
    ) -> Self:
        """
        Directly computes the minimal DFA recognizing strings whose length is
        between `min_length` and `max_length`, inclusive. To allow arbitrarily
        long words, the value `None` can be passed in for `max_length`.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        min_length : int, default: 0
            The minimum length of strings to be accepted by this DFA.
        max_length : Optional[int], default: None
            The maximum length of strings to be accepted by this DFA.
            If set to None, there is no maximum.
        symbols_to_count : Optional[AbstractSet[str]], default: None
            The input symbols to count towards the length of words to accepts.
            If set to None, counts all symbols.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        if symbols_to_count is None:
            symbols_to_count = input_symbols

        transitions = {}
        length_range = (
            range(min_length) if max_length is None else range(max_length + 1)
        )
        for prev_state in length_range:
            next_state = prev_state + 1
            transitions[prev_state] = {
                symbol: next_state if symbol in symbols_to_count else prev_state
                for symbol in input_symbols
            }
        last_state = len(transitions)
        transitions[last_state] = {symbol: last_state for symbol in input_symbols}
        final_states = (
            frozenset((last_state,))
            if max_length is None
            else frozenset(range(min_length, max_length + 1))
        )

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=final_states,
        )

    @classmethod
    def count_mod(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        k: int,
        *,
        remainders: Optional[AbstractSet[int]] = None,
        symbols_to_count: Optional[AbstractSet[str]] = None,
    ) -> Self:
        """
        Directly computes a DFA that counts given symbols and accepts all strings where
        the remainder of division by k is in the set of remainders given.
        The default value of remainders is {0} and all symbols are counted by default.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        k : int
            The number to divide the length by.
        remainders : Optional[AbstractSet[int]], default: None
            The remainders to accept. If set to None, defaults to {0}.
        symbols_to_count : Optional[AbstractSet[str]], default: None
            The input symbols to count towards the length of words to accepts.
            If set to None, counts all symbols.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        if k <= 0:
            raise ValueError("Integer must be positive")
        if symbols_to_count is None:
            symbols_to_count = input_symbols
        if remainders is None:
            remainders = {0}
        transitions = {
            i: {
                symbol: (i + 1) % k if symbol in symbols_to_count else i
                for symbol in input_symbols
            }
            for i in range(k)
        }

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states=remainders,
        )

    @classmethod
    def universal_language(cls: Type[Self], input_symbols: AbstractSet[str]) -> Self:
        """
        Directly computes the minimal DFA accepting all strings.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        return cls(
            states={0},
            input_symbols=input_symbols,
            transitions={0: {symbol: 0 for symbol in input_symbols}},
            initial_state=0,
            final_states={0},
            allow_partial=False,
        )

    @classmethod
    def empty_language(cls: Type[Self], input_symbols: AbstractSet[str]) -> Self:
        """
        Directly computes the minimal DFA rejecting all strings.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        return cls(
            states={0},
            input_symbols=input_symbols,
            transitions={0: {symbol: 0 for symbol in input_symbols}},
            initial_state=0,
            final_states=frozenset(),
            allow_partial=False,
        )

    @classmethod
    def nth_from_start(
        cls: Type[Self], input_symbols: AbstractSet[str], symbol: str, n: int
    ) -> Self:
        """
        Directly computes the minimal DFA which accepts all words whose `n`th
        character from the start is `symbol`, where `n` is a positive integer.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        symbol : str
            The target input symbol.
        n : int
            The position of the target input symbol.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        if n < 1:
            raise ValueError("Integer must be positive")
        if symbol not in input_symbols:
            raise exceptions.InvalidSymbolError(
                "Desired symbol is not in the set of input symbols"
            )
        if len(input_symbols) == 1:
            return cls.of_length(input_symbols, min_length=n)
        transitions = {i: {symbol: i + 1 for symbol in input_symbols} for i in range(n)}
        transitions[n - 1][symbol] = n + 1
        transitions[n] = {symbol: n for symbol in input_symbols}
        transitions[n + 1] = {symbol: n + 1 for symbol in input_symbols}

        return cls(
            states=frozenset(transitions.keys()),
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=0,
            final_states={n + 1},
            allow_partial=False,
        )

    @classmethod
    def nth_from_end(
        cls: Type[Self], input_symbols: AbstractSet[str], symbol: str, n: int
    ) -> Self:
        """
        Directly computes the minimal DFA which accepts all words whose `n`th
        character from the end is `symbol`, where `n` is a positive integer.

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        symbol : str
            The target input symbol.
        n : int
            The position of the target input symbol.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """
        if n < 1:
            raise ValueError("Integer must be positive")
        if symbol not in input_symbols:
            raise exceptions.InvalidSymbolError(
                "Desired symbol is not in the set of input symbols"
            )
        if len(input_symbols) == 1:
            return cls.of_length(input_symbols, min_length=n)
        # Consider the states to be labelled with bitstrings of size n The
        # bitstring represents how the current suffix in this state matches our
        # desired symbol A 1 means the character at this position is the desired
        # symbol and a 0 means it is not For transitions this is effectively
        # doubling the label value and then adding 1 if the desired symbol is
        # read Finally we trim the label to n bits with a modulo operation.
        state_count = 2**n

        return cls(
            states=frozenset(range(state_count)),
            input_symbols=input_symbols,
            transitions={
                state: {
                    sym: (
                        (2 * state + 1) % state_count
                        if symbol == sym
                        else (2 * state) % state_count
                    )
                    for sym in input_symbols
                }
                for state in range(state_count)
            },
            initial_state=0,
            final_states=frozenset(range(state_count // 2, state_count)),
            allow_partial=False,
        )

    @classmethod
    def from_finite_language(
        cls: Type[Self],
        input_symbols: AbstractSet[str],
        language: AbstractSet[str],
        as_partial: bool = True,
    ) -> Self:
        """
        Directly computes the minimal DFA accepting the finite language given as input.
        Uses the algorithm described in Finite-State Techniques by Mihov and Schulz,
        Chapter 10

        Parameters
        ----------
        input_symbols : AbstractSet[str]
            The set of input symbols to construct the DFA over.
        language : AbstractSet[str]
            The language to accept.
        as_partial : bool, default: True
            Whether or not to construct this as a partial DFA.

        Returns
        ------
        Self
            The DFA accepting the desired language.
        """

        SignatureT: TypeAlias = Tuple[bool, FrozenSet[Tuple[str, str]]]

        if not language:
            return cls.empty_language(input_symbols)

        transitions: Dict[DFAStateT, Dict[str, DFAStateT]] = {}
        back_map: Dict[str, Set[str]] = {"": set()}
        final_states: Set[str] = set()
        signatures_dict: Dict[SignatureT, str] = {}

        def compute_signature(state: str) -> SignatureT:
            """Computes signature for input state"""
            return (state in final_states, frozenset(transitions[state].items()))

        def longest_common_prefix_length(string_1: str, string_2: str) -> int:
            """Returns length of longest common prefix."""
            for i, (symbol_1, symbol_2) in enumerate(zip(string_1, string_2)):
                if symbol_1 != symbol_2:
                    return i

            return min(len(string_1), len(string_2))

        def add_to_trie(word: str) -> None:
            """Add word to the trie represented by transitions"""
            prefix = ""
            for symbol in word:
                next_prefix = prefix + symbol

                # Extend the trie only if necessary
                prefix_dict = transitions.setdefault(prefix, {})
                prefix_dict.setdefault(symbol, next_prefix)
                back_map.setdefault(next_prefix, set()).add(prefix)

                prefix = next_prefix

            # Mark the finished prefix as a final state
            transitions[prefix] = {}
            final_states.add(prefix)

        def compress(word: str, next_word: str) -> None:
            """Compress prefixes of word, newly added to the DFA"""

            # Compress along all prefixes that are _not_ prefixes of the next word
            lcp_len = longest_common_prefix_length(word, next_word)

            # Compression in order of reverse length
            for i in range(len(word), lcp_len, -1):
                prefix = word[:i]

                # Compute signature for comparison
                prefix_signature = compute_signature(prefix)
                identical_state = signatures_dict.get(prefix_signature)

                if identical_state is not None:
                    # If identical state exists, remove prefix since it's redundant
                    final_states.discard(prefix)
                    transitions.pop(prefix)

                    # Change transition for prefix
                    for parent_state in back_map[prefix]:
                        path = transitions[parent_state]
                        for symbol in path:
                            if path[symbol] == prefix:
                                path[symbol] = identical_state
                        back_map[identical_state].add(parent_state)

                        # No need to recompute signatures here, since we will
                        # recompute when handling parents in later iterations
                else:
                    signatures_dict[prefix_signature] = prefix

        # Construct initial trie
        prev_word, *rest_words = sorted(language)
        add_to_trie(prev_word)

        for curr_word in rest_words:
            # For each word, compress from the previous iteration and continue
            compress(prev_word, curr_word)
            add_to_trie(curr_word)
            prev_word = curr_word

        compress(prev_word, "")

        if as_partial:
            return cls(
                states=frozenset(transitions.keys()),
                input_symbols=input_symbols,
                transitions=transitions,
                initial_state="",
                final_states=final_states,
                allow_partial=as_partial,
            )

        return cls._to_complete(
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state="",
            final_states=final_states,
            trap_state=0,
        )

    @classmethod
    def from_nfa(
        cls: Type[Self],
        target_nfa: nfa.NFA,
        *,
        retain_names: bool = False,
        minify: bool = True,
    ) -> Self:
        """
        Initialize this DFA as one equivalent to the given NFA. Note
        that this usually returns a partial DFA by default.

        Parameters
        ----------
        target_nfa : NFA
            The NFA to construct an equivalent DFA for.
        retain_names : bool, default: False
            Whether or not to retain state names during processing.
        minify : bool, default: True
            Whether or not to minify the DFA resulting from the input NFA.

        Returns
        ------
        Self
            The DFA accepting the language of the input NFA.
        """

        def subset_function(current_states: FrozenSet[DFAStateT]) -> bool:
            return not current_states.isdisjoint(target_nfa.final_states)

        initial_state = frozenset(
            target_nfa._get_lambda_closures()[target_nfa.initial_state]
        )

        return cls._expand_dfa(
            subset_function,
            initial_state,
            target_nfa._iterate_through_symbol_path_pairs,
            target_nfa.input_symbols,
            retain_names=retain_names,
            minify=minify,
        )

    def iter_transitions(
        self,
    ) -> Generator[Tuple[DFAStateT, DFAStateT, str], None, None]:
        """
        Iterate over all transitions in the DFA. Each transition is a tuple
        of the form (from_state, to_state, symbol).

        Returns
        ------
        Generator[Tuple[DFAStateT, DFAStateT, str], None, None]
            The desired generator over the DFA transitions.
        """
        return (
            (from_, to_, symbol)
            for from_, lookup in self.transitions.items()
            for symbol, to_ in lookup.items()
        )

    def _get_input_path(
        self, input_str: str
    ) -> Tuple[List[Tuple[DFAStateT, DFAStateT, DFASymbolT]], bool]:
        """
        Calculate the path taken by input.

        Parameters
        ------
        input_str : str
            The input string to run on the DFA.

        Returns
        ------
        Tuple[List[Tuple[DFAStateT, DFAStateT, DFASymbolT], bool]]
            A list of all transitions taken in each step and a boolean
            indicating whether the DFA accepted the input.

        """

        state_history = list(self.read_input_stepwise(input_str, ignore_rejection=True))

        path = [
            (*state_pair, char)
            for state_pair, char in zip(pairwise(state_history), input_str)
        ]

        last_state = state_history[-1] if state_history else self.initial_state
        accepted = last_state in self.final_states

        return path, accepted



"""Miscellaneous utility functions and classes."""

from __future__ import annotations

import os
import pathlib
import random
import uuid
from collections import defaultdict, deque
from itertools import count, tee, zip_longest
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import networkx as nx
from frozendict import frozendict

# Optional imports for use with visual functionality
try:
    import pygraphviz as pgv
except ImportError:
    _visual_imports = False
else:
    _visual_imports = True


LayoutMethod = Literal["neato", "dot", "twopi", "circo", "fdp", "nop"]


def freeze_value(value: Any) -> Any:
    """
    A helper function to convert the given value / structure into a fully
    immutable one by recursively processing said structure and any of its
    members, freezing them as well
    """
    if isinstance(value, (str, int)):
        return value
    if isinstance(value, dict):
        return frozendict(
            {
                dict_key: freeze_value(dict_value)
                for dict_key, dict_value in value.items()
            }
        )
    if isinstance(value, set):
        return frozenset(freeze_value(element) for element in value)
    if isinstance(value, list):
        return tuple(freeze_value(element) for element in value)
    return value


def get_renaming_function(counter: count) -> Callable[[Any], int]:
    """
    A helper function that returns a renaming function to be used in the creation of
    other automata. The parameter counter should be an itertools count.
    This helper function will return the same distinct output taken from counter
    for each distinct input.
    """

    return defaultdict(counter.__next__).__getitem__


def create_unique_random_id() -> str:
    # To be able to set the random seed, took code from:
    # https://nathanielknight.ca/articles/consistent_random_uuids_in_python.html
    return str(
        uuid.UUID(bytes=bytes(random.getrandbits(8) for _ in range(16)), version=4)
    )


def create_graph(
    horizontal: bool = True,
    reverse_orientation: bool = False,
    fig_size: Union[Tuple[float, float], Tuple[float], None] = None,
    state_separation: float = 0.5,
) -> pgv.AGraph:
    """Creates and returns a graph object
    Args:
        - horizontal (bool, optional): Direction of node layout. Defaults
            to True.
        - reverse_orientation (bool, optional): Reverse direction of node
            layout. Defaults to False.
        - fig_size (tuple, optional): Figure size. Defaults to None.
        - state_separation (float, optional): Node distance. Defaults to 0.5.
    Returns:
        AGraph with the given configuration.
    """
    if not _visual_imports:
        raise ImportError(
            "Missing visualization packages; please install coloraide and pygraphviz."
        )

    # Defining the graph.
    graph = pgv.AGraph(strict=False, directed=True)

    if fig_size is not None:
        graph.graph_attr.update(size=", ".join(map(str, fig_size)))

    graph.graph_attr.update(ranksep=str(state_separation))

    if horizontal:
        rankdir = "RL" if reverse_orientation else "LR"
    else:
        rankdir = "BT" if reverse_orientation else "TB"

    graph.graph_attr.update(rankdir=rankdir)

    return graph


def save_graph(
    graph: pgv.AGraph,
    path: Union[str, os.PathLike],
) -> None:
    """Write `graph` to file given by `path`. PNG, SVG, etc.
    Returns the same graph."""

    save_path_final: pathlib.Path = pathlib.Path(path)

    format = save_path_final.suffix.split(".")[1] if save_path_final.suffix else None

    graph.draw(
        path=save_path_final,
        format=format,
    )


T = TypeVar("T")


class PartitionRefinement(Generic[T]):
    """Maintain and refine a partition of a set of items into subsets.
    Space usage for a partition of n items is O(n), and each refine operation
    takes time proportional to the size of its argument.

    Adapted from code by D. Eppstein:
    https://www.ics.uci.edu/~eppstein/PADS/PartitionRefinement.py
    """

    __slots__: Tuple[str, ...] = ("_sets", "_partition")

    _sets: Dict[int, Set[T]]
    _partition: Dict[T, int]

    def __init__(self, items: Iterable[T]) -> None:
        """Create a new partition refinement data structure for the given
        items. Initially, all items belong to the same subset.
        """
        S = set(items)
        self._sets = {id(S): S}
        self._partition = {x: id(S) for x in S}

    def get_set_by_id(self, id: int) -> Set[T]:
        """Return the set in the partition corresponding to id."""
        return self._sets[id]

    def get_set_ids(self) -> Iterable[int]:
        """Return set ids corresponding to the internal partition."""
        return self._sets.keys()

    def get_sets(self) -> Iterable[Set[T]]:
        """Return sets corresponding to the internal partition."""
        return self._sets.values()

    def refine(self, S: Iterable[T]) -> List[Tuple[int, int]]:
        """Refine each set A in the partition to the two sets
        A & S, A - S.  Return a list of pairs ids (id(A & S), id(A - S))
        for each changed set.  Within each pair, A & S will be
        a newly created set, while A - S will be a modified
        version of an existing set in the partition (retaining its old id).
        Not a generator because we need to perform the partition
        even if the caller doesn't iterate through the results.
        """
        hit = defaultdict(set)
        output = []

        for x in S:
            hit[self._partition[x]].add(x)

        for Aid, AintS in hit.items():
            A = self._sets[Aid]

            # Only need to check lengths, we already know AintS is a subset of A
            # by construction
            if len(AintS) < len(A):
                self._sets[id(AintS)] = AintS
                for x in AintS:
                    self._partition[x] = id(AintS)
                A -= AintS
                output.append((id(AintS), Aid))

        return output


def pairwise(iterable: Iterable[T], final_none: bool = False) -> Iterable[Tuple[T, T]]:
    """Based on https://docs.python.org/3/library/itertools.html#itertools.pairwise"""
    a, b = tee(iterable)
    next(b, None)

    if final_none:
        return zip_longest(a, b)

    return zip(a, b)


def get_reachable_nodes(
    G: nx.DiGraph,
    sources: Iterable[Any],
    reversed: bool = False,
) -> Set[Any]:
    """
    Return a set with all descendants (or predecessors if reversed is True)
    of the nodes in sources.

    Adapted from:
    https://networkx.org/documentation/stable/_modules/networkx/algorithms/traversal/breadth_first_search.html#generic_bfs_edges
    """
    seen = set(sources)
    get_neighbors = G.predecessors if reversed else G.neighbors
    work_deque = deque(sources)
    while work_deque:
        next_node = work_deque.popleft()
        for neighbor in get_neighbors(next_node):
            if neighbor in seen:
                continue

            seen.add(neighbor)
            work_deque.append(neighbor)

    return seen



"""Tests for automata pickle serialization"""

import pickle

import tests.test_fa as test_fa


class TestSerialization(test_fa.TestFA):
    """
    This tests verifies that the cycle FA -> .pkl -> FA works.
    The test only applies to DFA and NFAs as other classes do not implement equality
    """

    def test_serialize_dfa(self) -> None:
        """Should convert a DFA to pickle serialization and reads it back"""
        s = pickle.dumps(self.dfa)
        dfa = pickle.loads(s)
        self.assertEqual(self.dfa, dfa)

    def test_serialize_nfa(self) -> None:
        """Should convert a NFA to pickled representation and read it back"""
        s = pickle.dumps(self.nfa)
        nfa = pickle.loads(s)
        self.assertEqual(self.nfa, nfa)



"""Classes and functions for testing the behavior of PDAConfigurations."""

import tests.test_pda as test_pda
from automata.tm.configuration import TMConfiguration
from automata.tm.tape import TMTape


class TestTMConfiguration(test_pda.TestPDA):
    """A test class for testing configurations of pushdown automata."""

    def test_config_hashability(self) -> None:
        self.assertEqual(
            hash(
                TMConfiguration(
                    "q0", TMTape("01", blank_symbol=".", current_position=0)
                )
            ),
            hash(
                TMConfiguration(
                    "q0", TMTape("01", blank_symbol=".", current_position=0)
                )
            ),
        )

    def test_config_repr(self) -> None:
        """Should create proper string representation of PDA configuration."""
        config = TMConfiguration(
            "q0", TMTape("01", blank_symbol=".", current_position=0)
        )
        self.assertEqual(repr(config), "TMConfiguration('q0', TMTape('01', '.', 0))")



"""Classes and methods for working with deterministic pushdown automata."""

from typing import AbstractSet, Generator, List, Mapping, Optional, Set, Tuple, Union

import automata.base.exceptions as exceptions
import automata.pda.exceptions as pda_exceptions
import automata.pda.pda as pda
from automata.base.utils import pairwise
from automata.pda.configuration import PDAConfiguration
from automata.pda.stack import PDAStack

DPDAStateT = pda.PDAStateT

DPDAPathT = Mapping[str, Mapping[str, Tuple[DPDAStateT, Union[str, Tuple[str, ...]]]]]
DPDATransitionsT = Mapping[DPDAStateT, DPDAPathT]


class DPDA(pda.PDA):
    """
    The `DPDA` class is a subclass of `PDA` and represents a deterministic pushdown
    automaton.

    Parameters
    ----------
    states: AbstractSet[DPDAStateT]
        A set of the DPDA's valid states
    input_symbols: AbstractSet[str]
        Set of the DPDA's valid input symbols, each of which is a singleton
        string.
    stack_symbols: AbstractSet[str]
        Set of the DPDA's valid stack symbols, each of which is a singleton
        string.
    transitions: DPDATransitionsT
        A dict consisting of the transitions for each state; see the
        example below for the exact syntax
    initial_state: DPDAStateT
        The name of the initial state for this DPDA.
    initial_stack_symbol: str
        The name of the initial symbol on the stack for this DPDA.
    final_states: AbstractSet[DPDAStateT]
        A set of final states for this DPDA.
    acceptance_mode: pda.PDAAcceptanceModeT, default: "both"
        A string defining whether this DPDA accepts by
        `'final_state'`, `'empty_stack'`, or `'both'`.

    Example
    ----------
        from automata.pda.dpda import DPDA
        # DPDA which which matches zero or more 'a's, followed by the same
        # number of 'b's (accepting by final state)
        dpda = DPDA(
            states={'q0', 'q1', 'q2', 'q3'},
            input_symbols={'a', 'b'},
            stack_symbols={'0', '1'},
            transitions={
                'q0': {
                    'a': {'0': ('q1', ('1', '0'))}  # push '1' to stack
                },
                'q1': {
                    'a': {'1': ('q1', ('1', '1'))},  # push '1' to stack
                    'b': {'1': ('q2', '')}  # pop from stack
                },
                'q2': {
                    'b': {'1': ('q2', '')},  # pop from stack
                    '': {'0': ('q3', ('0',))}  # no change to stack
                }
            },
            initial_state='q0',
            initial_stack_symbol='0',
            final_states={'q3'},
            acceptance_mode='final_state'
        )
    """

    __slots__ = (
        "states",
        "input_symbols",
        "stack_symbols",
        "transitions",
        "initial_state",
        "initial_stack_symbol",
        "final_states",
        "acceptance_mode",
    )

    def __init__(
        self,
        *,
        states: AbstractSet[DPDAStateT],
        input_symbols: AbstractSet[str],
        stack_symbols: AbstractSet[str],
        transitions: DPDATransitionsT,
        initial_state: DPDAStateT,
        initial_stack_symbol: str,
        final_states: AbstractSet[DPDAStateT],
        acceptance_mode: pda.PDAAcceptanceModeT = "both",
    ) -> None:
        """Initialize a complete DPDA."""
        super().__init__(
            states=states,
            input_symbols=input_symbols,
            stack_symbols=stack_symbols,
            transitions=transitions,
            initial_state=initial_state,
            initial_stack_symbol=initial_stack_symbol,
            final_states=final_states,
            acceptance_mode=acceptance_mode,
        )

    def iter_transitions(
        self,
    ) -> Generator[Tuple[DPDAStateT, DPDAStateT, Tuple[str, str, str]], None, None]:
        return (
            (from_, to_, (input_symbol, stack_symbol, "".join(stack_push)))
            for from_, input_lookup in self.transitions.items()
            for input_symbol, stack_lookup in input_lookup.items()
            for stack_symbol, (to_, stack_push) in stack_lookup.items()
        )

    def _validate_transition_invalid_symbols(
        self, start_state: DPDAStateT, paths: DPDATransitionsT
    ) -> None:
        """Raise an error if transition symbols are invalid."""
        for input_symbol, symbol_paths in paths.items():
            self._validate_transition_invalid_input_symbols(start_state, input_symbol)
            for stack_symbol in symbol_paths:
                self._validate_transition_isolated_lambda_transitions(
                    start_state, input_symbol, stack_symbol
                )
                self._validate_transition_invalid_stack_symbols(
                    start_state, stack_symbol
                )

    def _validate_transition_lambda_transition_sibling(
        self, start_state: DPDAStateT, sib_path: DPDAPathT
    ) -> None:
        """Check the given sibling path for adjacent lambda transitions."""
        for other_stack_symbol in sib_path:
            if other_stack_symbol in self.transitions[start_state][""]:
                raise pda_exceptions.NondeterminismError(
                    "A symbol transition is adjacent to a "
                    "lambda transition for this DPDA."
                )

    def _validate_transition_isolated_lambda_transitions(
        self, start_state: DPDAStateT, input_symbol: str, stack_symbol: str
    ) -> None:
        """Raise an error if a lambda transition has no sibling transitions."""
        if input_symbol == "":
            sib_transitions = self.transitions[start_state]
            for sib_input_symbol, sib_path in sib_transitions.items():
                if sib_input_symbol != "":
                    self._validate_transition_lambda_transition_sibling(
                        start_state, sib_path
                    )

    def _get_transition(
        self, state: DPDAStateT, input_symbol: str, stack_symbol: str
    ) -> Optional[DPDAPathT]:
        """Get the transiton tuple for the given state and symbols."""
        if (
            state in self.transitions
            and input_symbol in self.transitions[state]
            and stack_symbol in self.transitions[state][input_symbol]
        ):
            return (input_symbol,) + self.transitions[state][input_symbol][stack_symbol]
        else:
            return None

    def _check_for_input_rejection(
        self, current_configuration: PDAConfiguration
    ) -> None:
        """Raise an error if the given config indicates rejected input."""
        if not self._has_accepted(current_configuration):
            raise exceptions.RejectionException(
                "the DPDA stopped in a non-accepting configuration ({}, {})".format(
                    current_configuration.state, current_configuration.stack
                )
            )

    def _get_next_configuration(self, old_config: PDAConfiguration) -> PDAConfiguration:
        """Advance to the next configuration."""
        transitions: Set[Optional[DPDAPathT]] = set()
        if old_config.remaining_input:
            transitions.add(
                self._get_transition(
                    old_config.state,
                    old_config.remaining_input[0],
                    old_config.stack.top(),
                )
            )
        transitions.add(
            self._get_transition(old_config.state, "", old_config.stack.top())
        )
        if None in transitions:
            transitions.remove(None)
        if len(transitions) == 0:
            raise exceptions.RejectionException(
                "The automaton entered a configuration for which no "
                "transition is defined ({}, {}, {})".format(
                    old_config.state,
                    old_config.remaining_input[0],
                    old_config.stack.top(),
                )
            )
        input_symbol, new_state, new_stack_top = transitions.pop()  # type: ignore
        remaining_input = old_config.remaining_input
        if input_symbol:
            remaining_input = remaining_input[1:]
        new_config = PDAConfiguration(
            new_state,
            remaining_input,
            self._replace_stack_top(old_config.stack, new_stack_top),
        )
        return new_config

    def _get_input_path(
        self, input_str: str
    ) -> Tuple[List[Tuple[PDAConfiguration, PDAConfiguration]], bool]:
        """
        Calculate the path taken by input.

        Args:
            input_str (str): The input string to run on the DPDA.

        Returns:
            Tuple[List[Tuple[PDAConfiguration, PDAConfiguration]], bool]: A list
            of all transitions taken in each step and a boolean indicating
            whether the DPDA accepted the input.

        """

        state_history = list(self.read_input_stepwise(input_str))

        path = list(pairwise(state_history))

        last_state = state_history[-1] if state_history else self.initial_state
        accepted = last_state in self.final_states

        return path, accepted

    def read_input_stepwise(
        self, input_str: str
    ) -> Generator[PDAConfiguration, None, None]:
        """
        Return a generator that yields the configuration of this DPDA at each
        step while reading input.

        Parameters
        ----------
        input_str : str
            The input string to read.

        Yields
        ------
        Generator[PDAConfiguration, None, None]
            A generator that yields the current configuration of
            the DPDA after each step of reading input.

        Raises
        ------
        RejectionException
            Raised if this DPDA does not accept the input string.
        """
        current_configuration = PDAConfiguration(
            self.initial_state, input_str, PDAStack([self.initial_stack_symbol])
        )

        yield current_configuration
        while current_configuration.remaining_input or self._has_lambda_transition(
            current_configuration.state, current_configuration.stack.top()
        ):
            current_configuration = self._get_next_configuration(current_configuration)
            yield current_configuration
            if self._has_accepted(current_configuration):
                return
        self._check_for_input_rejection(current_configuration)



"""Classes and methods for working with generalized non-deterministic finite
automata."""

from __future__ import annotations

from itertools import product
from typing import (
    AbstractSet,
    Dict,
    Generator,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    cast,
)

from typing_extensions import NoReturn, Self

import automata.base.exceptions as exceptions
import automata.fa.dfa as dfa
import automata.fa.fa as fa
import automata.fa.nfa as nfa
import automata.regex.regex as re

GNFAStateT = fa.AutomatonStateT

GNFAPathT = Mapping[GNFAStateT, Optional[str]]
GNFATransitionsT = Mapping[GNFAStateT, GNFAPathT]


class GNFA(fa.FA):
    """
    The `GNFA` class is a subclass of `FA` and represents a generalized
    nondeterministic finite automaton.

    Its main usage is for conversion of DFAs and NFAs to regular expressions. Note
    that because of this, the `GNFA` doesn't support any binary operators or reading
    input (e.g. `read_input_stepwise`). Every GNFA can be rendered natively inside of-
    a Jupyter notebook (automatically calling `show_diagram` without any arguments)
    if installed with the `visual` optional dependency. Note that `input_str`
    cannot be set as an argument to `show_diagram`, as the `GNFA` does not read input.

    Except for `initial_state` and `final_state`, one transition goes from every state
    to every other state, and also from each state to itself. To accommodate
    this, transitions can be regular expressions and `None`, in addition to
    normal input symbols.

    Parameters
    ----------
    states : AbstractSet[GNFAStateT]
        Set of the GNFA's valid states.
    input_symbols : AbstractSet[str]
        Set of the GNFA's valid input symbols, each of which is a singleton
        string.
    transitions : Mapping[GNFAStateT, Mapping[GNFAStateT, Optional[str]]]
        A dict with the transitions for each state, except `final_state`.
        Each key is a state name and each value is dict which maps a state
        (the key) to the transition expression (the value) or None.
        The transition expression is a regular expression (string) over
        `input_symbols`, and the following symbols only: `*`, `|`, `?`, `()`.

        This is a subset of the standard regular expressions using this package.
    initial_state : GNFAStateT
        The initial state for this GNFA. Has transitions going to every
        other state, but no transitions coming in from any other state.
    final_state : GNFAStateT
        A single final state for this GNFA. Has transitions coming in from every
        other state, but no transitions going to any other state.
        Must be different from the `initial_state`.
    """

    # Add __dict__ to deal with inheritance issue and the final_states attribute.
    __slots__ = (
        "states",
        "input_symbols",
        "transitions",
        "initial_state",
        "final_state",
        "__dict__",
    )

    final_state: GNFAStateT

    def __init__(
        self,
        *,
        states: AbstractSet[GNFAStateT],
        input_symbols: AbstractSet[str],
        transitions: GNFATransitionsT,
        initial_state: GNFAStateT,
        final_state: GNFAStateT,
    ) -> None:
        """Initialize a complete GNFA."""
        super(fa.FA, self).__init__(
            states=states,
            input_symbols=input_symbols,
            transitions=transitions,
            initial_state=initial_state,
            final_state=final_state,
            final_states={final_state},
        )

    # GNFA should NOT create the lambda closures via NFA.__post_init__()
    def __post_init__(self) -> None:
        self.validate()

    @classmethod
    def from_dfa(cls: Type[Self], target_dfa: dfa.DFA) -> Self:
        """
        Initialize this GNFA as one equivalent to the given DFA.

        Parameters
        ----------
        target_dfa : DFA
            The DFA to construct an equivalent GNFA for.

        Returns
        ------
        Self
            The GNFA accepting the language of the input DFA.
        """
        new_gnfa_transitions = dict()
        gnfa_states = set(target_dfa.states)

        for state in target_dfa.states:
            gnfa_transitions: Dict[GNFAStateT, Optional[str]] = dict()
            if state in target_dfa.transitions:
                for input_symbol, to_state in target_dfa.transitions[state].items():
                    if to_state in gnfa_transitions.keys():
                        gnfa_transitions[to_state] = (
                            f"{gnfa_transitions[to_state]}|{input_symbol}"
                        )
                    else:
                        gnfa_transitions[to_state] = input_symbol
                new_gnfa_transitions[state] = gnfa_transitions
            else:
                new_gnfa_transitions[state] = dict()

        new_initial_state = GNFA._add_new_state(gnfa_states)
        new_final_state = GNFA._add_new_state(gnfa_states, new_initial_state)

        new_gnfa_transitions[new_initial_state] = {target_dfa.initial_state: ""}

        for state in target_dfa.final_states:
            new_gnfa_transitions[state][new_final_state] = ""

        for state in gnfa_states - {new_final_state}:
            if gnfa_states - new_gnfa_transitions[state].keys():
                for leftover_state in gnfa_states - new_gnfa_transitions[state].keys():
                    if leftover_state is not new_initial_state:
                        new_gnfa_transitions[state][leftover_state] = None

        return cls(
            states=gnfa_states,
            input_symbols=target_dfa.input_symbols,
            transitions=new_gnfa_transitions,
            initial_state=new_initial_state,
            final_state=new_final_state,
        )

    @classmethod
    def from_nfa(cls: Type[Self], target_nfa: nfa.NFA) -> Self:
        """
        Initialize this GNFA as one equivalent to the given NFA.

        Parameters
        ----------
        target_nfa : NFA
            The NFA to construct an equivalent GNFA for.

        Returns
        ------
        Self
            The GNFA accepting the language of the input NFA.
        """
        new_gnfa_transitions: Dict[GNFAStateT, Dict[GNFAStateT, Optional[str]]] = dict()
        gnfa_states = set(target_nfa.states)

        for state in target_nfa.states:
            gnfa_transitions: Dict[GNFAStateT, str] = dict()
            if state in target_nfa.transitions:
                for input_symbol, to_states in target_nfa.transitions[state].items():
                    for to_state in to_states:
                        if to_state in gnfa_transitions.keys():
                            if gnfa_transitions[to_state] == "" and input_symbol != "":
                                gnfa_transitions[to_state] = f"{input_symbol}?"
                            elif (
                                gnfa_transitions[to_state] != "" and input_symbol == ""
                            ):
                                if cls._isbracket_req(gnfa_transitions[to_state]):
                                    gnfa_transitions[to_state] = (
                                        f"({gnfa_transitions[to_state]})?"
                                    )
                                else:
                                    gnfa_transitions[to_state] = (
                                        f"{gnfa_transitions[to_state]}?"
                                    )
                            else:
                                gnfa_transitions[to_state] = (
                                    f"{gnfa_transitions[to_state]}|{input_symbol}"
                                )
                        else:
                            gnfa_transitions[to_state] = input_symbol
                new_gnfa_transitions[state] = cast(
                    Dict[GNFAStateT, Optional[str]], gnfa_transitions
                )
            else:
                new_gnfa_transitions[state] = dict()

        new_initial_state = GNFA._add_new_state(gnfa_states)
        new_final_state = GNFA._add_new_state(gnfa_states, new_initial_state)

        new_gnfa_transitions[new_initial_state] = {target_nfa.initial_state: ""}

        for state in target_nfa.final_states:
            new_gnfa_transitions[state][new_final_state] = ""

        for state in gnfa_states - {new_final_state}:
            if gnfa_states - new_gnfa_transitions[state].keys():
                for leftover_state in gnfa_states - new_gnfa_transitions[state].keys():
                    if leftover_state is not new_initial_state:
                        new_gnfa_transitions[state][leftover_state] = None

        return cls(
            states=gnfa_states,
            input_symbols=target_nfa.input_symbols,
            transitions=new_gnfa_transitions,
            initial_state=new_initial_state,
            final_state=new_final_state,
        )

    def _validate_transition_invalid_symbols(
        self, start_state: GNFAStateT, paths: GNFAPathT
    ) -> None:
        """Raise an error if transition symbols are invalid."""
        check = self.input_symbols | {"*", "|", "(", ")", "?"}

        for regex in paths.values():
            if regex is not None and (
                set(regex) - check and regex != "" or not re._validate(regex)
            ):
                raise exceptions.InvalidRegexError(
                    "state {} has invalid transition expression {}".format(
                        start_state, regex
                    )
                )

    def _validate_transition_end_states(
        self, start_state: GNFAStateT, paths: GNFAPathT
    ) -> None:
        """Raise an error if transition end states are invalid or missing"""
        if start_state == self.final_state:
            if len(paths) != 0:
                raise exceptions.InvalidStateError(
                    "No transitions should be defined for final state {}".format(
                        start_state
                    )
                )
        elif (
            start_state == self.initial_state
            and self.states - paths.keys() - {self.initial_state} != set()
        ):
            raise exceptions.MissingStateError(
                "state {} does not have transitions defined for states {}".format(
                    start_state, str(self.states - paths.keys() - {self.initial_state})
                )
            )
        elif (
            start_state != self.initial_state
            and self.states - paths.keys() - {self.initial_state} != set()
        ):
            raise exceptions.MissingStateError(
                "state {} does not have transitions defined for states {}".format(
                    start_state, str(self.states - paths.keys() - {self.initial_state})
                )
            )
        for end_state in paths.keys():
            if end_state not in self.states:
                raise exceptions.InvalidStateError(
                    "end state {} for transition on {} is not valid".format(
                        end_state, start_state
                    )
                )

    def _validate_final_state(self) -> None:
        """Raise an error if the initial state is invalid."""
        if self.final_state not in self.states:
            raise exceptions.InvalidStateError(
                "{} is not a valid final state".format(self.final_state)
            )

    def validate(self) -> None:
        """
        Raises an exception if this automaton is not internally consistent.

        Raises
        ------
        InvalidStateError
            If this GNFA has invalid states in the transition dictionary.
        MissingStateError
            If this GNFA has states missing from the transition dictionary.
        InvalidRegexError
            If this GNFA has invalid regex in the transition dictionary.
        """
        self._validate_initial_state()
        self._validate_final_state()
        for start_state, paths in self.transitions.items():
            self._validate_transition_invalid_symbols(start_state, paths)
            self._validate_transition_end_states(start_state, paths)
        self._validate_initial_state_transitions()

    @staticmethod
    def _isbracket_req(regex: str) -> bool:
        bracket_open = 0
        for symbol in regex:
            if symbol == "(":
                bracket_open += 1
            elif symbol == ")":
                bracket_open -= 1
            if bracket_open == 0 and symbol == "|":
                return True

        return False

    @staticmethod
    def _find_min_connected_node(
        states: Set[GNFAStateT],
        transitions: GNFATransitionsT,
        initial_state: GNFAStateT,
        final_state: GNFAStateT,
    ) -> GNFAStateT:
        state_set = states - {initial_state, final_state}
        state_degree = dict.fromkeys(state_set, 0)

        for state in states - {final_state}:
            for to_state, label in transitions[state].items():
                if label is not None:
                    if state != initial_state:
                        state_degree[state] += 1
                    if to_state != final_state:
                        state_degree[to_state] += 1

        return min(state_degree, key=lambda x: state_degree.get(x, -1))

    def to_regex(self) -> str:
        """
        Convert GNFA to regular expression.

        Returns
        ------
        str
            A regular expression equivalent to the input GNFA.
        """
        new_states = set(self.states)
        new_transitions = {
            state: dict(paths) for state, paths in self.transitions.items()
        }

        while len(new_states) > 2:
            q_rip = self._find_min_connected_node(
                new_states, new_transitions, self.initial_state, self.final_state
            )
            new_states.remove(q_rip)
            for q_i, q_j in product(
                new_states - {self.final_state}, new_states - {self.initial_state}
            ):
                r1 = new_transitions[q_i][q_rip]
                r2 = new_transitions[q_rip][q_rip]
                r3 = new_transitions[q_rip][q_j]
                r4 = new_transitions[q_i][q_j]

                if r1 is None or r3 is None:
                    new_transitions[q_i][q_j] = r4
                else:
                    # check if putting brackets around r1 is necessary
                    if self._isbracket_req(r1):
                        r1 = f"({r1})"

                    if r2 is None:
                        r2 = ""
                    elif len(r2) == 1:
                        r2 = f"{r2}*"
                    else:
                        r2 = f"({r2})*"

                    if self._isbracket_req(r3):
                        r3 = f"({r3})"

                    if r4 is None:
                        r4 = ""
                    elif self._isbracket_req(r4):
                        r4 = f"|({r4})"
                    elif r4 == "":
                        r4 = "?"
                    else:
                        r4 = f"|{r4}"

                    if r4 == "?" and len(r1) + len(r2) + len(r3) > 1:
                        new_transitions[q_i][q_j] = f"({r1}{r2}{r3}){r4}"
                    else:
                        new_transitions[q_i][q_j] = f"{r1}{r2}{r3}{r4}"

            del new_transitions[q_rip]
            for state in new_states - {self.final_state}:
                del new_transitions[state][q_rip]

        return new_transitions[self.initial_state][self.final_state]

    def read_input_stepwise(self, input_str: str) -> NoReturn:
        # No docstring because this is a dummy implementation
        raise NotImplementedError

    def iter_transitions(
        self,
    ) -> Generator[Tuple[GNFAStateT, GNFAStateT, str], None, None]:
        """
        Iterate over all transitions in the GNFA. Each transition is a tuple
        of the form (from_state, to_state, symbol).

        Returns
        ------
        Generator[Tuple[GNFAStateT, GNFAStateT, str], None, None]
            The desired generator over the GNFA transitions.
        """
        return (
            (from_, to_, symbol)
            for from_, lookup in self.transitions.items()
            for to_, symbol in lookup.items()
            if symbol is not None
        )

    def _get_input_path(self, input_str: str) -> NoReturn:
        raise NotImplementedError(
            f"_get_input_path is not implemented for {self.__class__}"
        )



"""
Classes and methods for lexing expressions into lists of tokens.

Adapted from code in the tdparser library: https://github.com/rbarrois/tdparser
"""

import abc
import re
from typing import (
    AbstractSet,
    Callable,
    FrozenSet,
    Generator,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from typing_extensions import Self

import automata.base.exceptions as exceptions

ResultT = TypeVar("ResultT")


class Token(Generic[ResultT], metaclass=abc.ABCMeta):
    """Base class for tokens."""

    __slots__: Tuple[str, ...] = ("text",)

    text: str

    def __init__(self, text: str) -> None:
        self.text = text

    @classmethod
    def from_match(cls: Type[Self], match: re.Match) -> Self:
        return cls(match.group())

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.text}>"


TokenFactoryT = Callable[[re.Match], Token[ResultT]]


class TokenRegistry(Generic[ResultT]):
    """Registry holding token rules."""

    __slots__: Tuple[str, ...] = ("_tokens",)

    _tokens: List[Tuple[TokenFactoryT, re.Pattern]]

    def __init__(self) -> None:
        self._tokens = []

    def register(self, token_factory_fn: TokenFactoryT, token_regex: str) -> None:
        """
        Register a token that can be produced by token_factory_fn (a function
        taking in a regex match and returning the final token) and recognized by the
        token_regex pattern.
        """
        self._tokens.append((token_factory_fn, re.compile(token_regex)))

    def matching_tokens(
        self, text: str, start: int
    ) -> Generator[Tuple[TokenFactoryT, re.Match], None, None]:
        """Retrieve all token definitions matching text starting at start."""

        for token_factory_fn, regexp in self._tokens:
            match = regexp.match(text, pos=start)
            if match:
                yield (token_factory_fn, match)

    def get_token(
        self, text: str, start: int = 0
    ) -> Optional[Tuple[TokenFactoryT, re.Match]]:
        """
        Retrieve the next token from some text. Computes the best match by
        length. Returns None if there is no match in the token registry.
        """
        best_token_match = None
        best_match = None

        for token_factory_fn, match in self.matching_tokens(text, start):
            if not best_match or best_match.end() < match.end():
                best_token_match = (token_factory_fn, match)
                best_match = match

        return best_token_match

    def __len__(self) -> int:
        return len(self._tokens)


class Lexer(Generic[ResultT]):
    """
    The core lexer. First, tokens are registered with their factory functions and regex
    patterns. The lexer can then take in a string and splits it into a list of token
    classes (in infix ordering) matching the regex patterns.
    """

    __slots__: Tuple[str, ...] = ("tokens", "blank_chars")

    tokens: TokenRegistry[ResultT]
    blank_chars: FrozenSet[str]

    def __init__(self, blank_chars: Optional[AbstractSet] = None) -> None:
        self.tokens = TokenRegistry()
        self.blank_chars = (
            frozenset((" ", "\t")) if blank_chars is None else frozenset(blank_chars)
        )

    def register_token(self, token_factory_fn: TokenFactoryT, token_regex: str) -> None:
        """
        Register a token class. The token_factory_fn must taken in a
        match object and return an instance of the desired token, and token_regex
        is used by the lexer to match tokens in the input text.
        """

        self.tokens.register(token_factory_fn, token_regex)

    def lex(self, text: str) -> List[Token[ResultT]]:
        """Split text into a list of tokens in infix notation."""

        pos = 0
        res = []

        while pos < len(text):
            token_match = self.tokens.get_token(text, start=pos)
            if token_match is not None:
                token_factory_fn, match = token_match
                res.append(token_factory_fn(match))
                pos += len(match.group())
            elif text[pos] in self.blank_chars:
                pos += 1
            else:
                raise exceptions.LexerError(
                    f"Invalid character '{text[pos]}' in '{text}'", position=pos
                )

        return res



"""Classes and functions for testing the behavior of GNFAs."""

import os
import tempfile
from unittest.mock import MagicMock, patch

import automata.base.exceptions as exceptions
import tests.test_fa as test_fa
from automata.fa.dfa import DFA
from automata.fa.gnfa import GNFA
from automata.fa.nfa import NFA


class TestGNFA(test_fa.TestFA):
    """A test class for testing generalized nondeterministic finite automata."""

    temp_dir_path = tempfile.gettempdir()

    def test_methods_not_implemented(self) -> None:
        """Should raise NotImplementedError when calling non-implemented methods."""
        abstract_methods = {
            "_get_input_path": (GNFA, ""),
            "read_input_stepwise": (GNFA, ""),
        }
        for method_name, method_args in abstract_methods.items():
            with self.assertRaises(NotImplementedError):
                getattr(GNFA, method_name)(*method_args)

    def test_init_gnfa(self) -> None:
        """Should copy GNFA if passed into NFA constructor."""
        new_gnfa = self.gnfa.copy()
        self.assertIsNot(new_gnfa, self.gnfa)

    def test_init_nfa_missing_formal_params(self) -> None:
        """Should raise an error if formal NFA parameters are missing."""
        with self.assertRaises(TypeError):
            GNFA(  # type: ignore
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                initial_state="q0",
                final_state="q1",
            )

    def test_copy_gnfa(self) -> None:
        """Should create exact copy of NFA if copy() method is called."""
        new_gnfa = self.gnfa.copy()
        self.assertIsNot(new_gnfa, self.gnfa)

    def test_gnfa_immutable_attr_set(self) -> None:
        with self.assertRaises(AttributeError):
            self.gnfa.states = {}  # type: ignore

    def test_gnfa_immutable_attr_del(self) -> None:
        with self.assertRaises(AttributeError):
            del self.gnfa.states

    def test_init_dfa(self) -> None:
        """Should convert DFA to GNFA if passed into GNFA constructor."""
        gnfa = GNFA.from_dfa(self.dfa)
        self.assertEqual(gnfa.states, {0, 1, "q2", "q0", "q1"})
        self.assertEqual(gnfa.input_symbols, {"0", "1"})
        self.assertEqual(
            gnfa.transitions,
            {
                "q0": {"q0": "0", "q1": "1", 1: None, "q2": None},
                "q1": {"q0": "0", "q2": "1", 1: "", "q1": None},
                "q2": {"q2": "0", "q1": "1", 1: None, "q0": None},
                0: {"q0": "", 1: None, "q1": None, "q2": None},
            },
        )
        self.assertEqual(gnfa.initial_state, 0)

    def test_init_nfa(self) -> None:
        """Should convert NFA to GNFA if passed into GNFA constructor."""
        gnfa = GNFA.from_nfa(self.nfa)
        self.assertEqual(gnfa.states, {0, 1, "q0", "q1", "q2"})
        self.assertEqual(gnfa.input_symbols, {"b", "a"})
        self.assertEqual(
            gnfa.transitions,
            {
                0: {1: None, "q0": "", "q1": None, "q2": None},
                "q0": {1: None, "q0": None, "q1": "a", "q2": None},
                "q1": {1: "", "q0": None, "q1": "a", "q2": ""},
                "q2": {1: None, "q0": "b", "q1": None, "q2": None},
            },
        )
        self.assertEqual(gnfa.initial_state, 0)

    @patch("automata.fa.gnfa.GNFA.validate")
    def test_init_validation(self, validate: MagicMock) -> None:
        """Should validate NFA when initialized."""
        self.gnfa.copy()
        validate.assert_called_once_with()

    def test_validate_invalid_symbol(self) -> None:
        """Should raise error if a transition references an invalid symbol."""
        with self.assertRaises(exceptions.InvalidRegexError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "c", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

    def test_validate_invalid_state(self) -> None:
        """Should raise error if a transition references an invalid state."""
        with self.assertRaises(exceptions.InvalidStateError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q3": "a", "q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

    def test_validate_invalid_initial_state(self) -> None:
        """Should raise error if the initial state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q3",
                final_state="q_f",
            )

    def test_validate_initial_state_transitions(self) -> None:
        """Should raise error if the initial state has no transitions."""
        with self.assertRaises(exceptions.MissingStateError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

    def test_validate_invalid_final_state(self) -> None:
        """Should raise error if the final state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q3",
            )

    def test_validate_final_state_transition(self) -> None:
        """Should raise error if there are transitions from final state"""
        with self.assertRaises(exceptions.InvalidStateError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                    "q_f": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

    def test_validate_missing_state(self) -> None:
        """Should raise an error if some transitions are missing."""
        with self.assertRaises(exceptions.MissingStateError):
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

    def test_validate_incomplete_transitions(self) -> None:
        """
        Should raise error if transitions from (except final state)
        and to (except initial state) every state is missing.
        """
        with self.assertRaises(exceptions.MissingStateError):
            # del gnfa.transitions['q1']['q0']
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": ""},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

        with self.assertRaises(exceptions.MissingStateError):
            # del gnfa.transitions['q_in']['q_f']
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {"q0": "", "q2": None, "q1": None},
                },
                initial_state="q_in",
                final_state="q_f",
            )

        with self.assertRaises(exceptions.InvalidStateError):
            # gnfa.transitions['q_in']['q5'] = {}
            GNFA(
                states={"q_in", "q_f", "q0", "q1", "q2"},
                input_symbols={"a", "b"},
                transitions={
                    "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                    "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                    "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                    "q_in": {
                        "q0": "",
                        "q_f": None,
                        "q2": None,
                        "q1": None,
                        "q5": {},  # type: ignore
                    },
                },
                initial_state="q_in",
                final_state="q_f",
            )

    def test_from_dfa(self) -> None:
        """
        Check if GNFA is generated properly from DFA
        """

        dfa = DFA(
            states={0, 1, 2, 4},
            input_symbols={"a", "b"},
            initial_state=0,
            final_states={4},
            transitions={
                0: {"a": 1, "b": 2},
                1: {"a": 2, "b": 2},
                2: {"b": 4},
                4: {},
            },
            allow_partial=True,
        )

        gnfa = GNFA.from_dfa(dfa)
        gnfa2 = GNFA(
            states={0, 1, 2, 3, 4, 5},
            input_symbols={"a", "b"},
            initial_state=3,
            final_state=5,
            transitions={
                0: {1: "a", 2: "b", 0: None, 4: None, 5: None},
                1: {2: "a|b", 0: None, 1: None, 4: None, 5: None},
                2: {4: "b", 0: None, 1: None, 2: None, 5: None},
                3: {0: "", 1: None, 2: None, 4: None, 5: None},
                4: {5: "", 0: None, 1: None, 2: None, 4: None},
            },
        )

        self.assertEqual(gnfa.input_parameters, gnfa2.input_parameters)

    def test_from_dfa_single_state(self) -> None:
        nfa = NFA.from_regex("")
        dfa = DFA.from_nfa(nfa)
        gnfa = GNFA.from_dfa(dfa)

        gnfa2 = GNFA(
            states={0, 1, "{0}"},
            input_symbols=set(),
            initial_state=0,
            final_state=1,
            transitions={"{0}": {1: "", "{0}": None}, 0: {"{0}": "", 1: None}},
        )

        self.assertEqual(gnfa.to_regex(), gnfa2.to_regex())

    def test_from_nfa_single_state(self) -> None:
        nfa = NFA.from_regex("")
        gnfa = GNFA.from_nfa(nfa)

        gnfa2 = GNFA(
            states={0, 1, 2},
            input_symbols=set(),
            initial_state=1,
            final_state=2,
            transitions={0: {2: "", 0: None}, 1: {0: "", 2: None}},
        )

        self.assertEqual(gnfa.to_regex(), gnfa2.to_regex())

    def test_from_nfa(self) -> None:
        """Should convert NFA to GNFA properly"""

        nfa = NFA(
            states={0, 1, 2, 4},
            input_symbols={"a", "b"},
            transitions={
                0: {"a": {1}, "b": {1}, "": {1}},
                1: {"a": {1, 2}, "": {2, 4}},
                2: {"": {0}, "b": {0, 4}},
            },
            initial_state=0,
            final_states={4},
        )

        gnfa = GNFA.from_nfa(nfa)

        gnfa2 = GNFA(
            states={0, 1, 2, 3, 4, 5},
            initial_state=3,
            final_state=5,
            input_symbols={"a", "b"},
            transitions={
                0: {1: "(a|b)?", 0: None, 2: None, 4: None, 5: None},
                1: {1: "a", 2: "a?", 4: "", 0: None, 5: None},
                2: {0: "b?", 4: "b", 1: None, 2: None, 5: None},
                4: {5: "", 0: None, 1: None, 2: None, 4: None},
                3: {0: "", 1: None, 2: None, 4: None, 5: None},
            },
        )

        self.assertEqual(gnfa.input_parameters, gnfa2.input_parameters)

    def test_to_regex(self) -> None:
        """
        We generate GNFA from DFA then convert it to regex
        then generate NFA from regex (already tested method)
        and check for equivalence of NFA and previous DFA
        """
        regex_strings = [
            "a*",
            "aa*b|bba*|(cc*)(bb+)",
            "a(aaa*bbcd|abbcd)d*|aa*bb(dcc*|(d|c)b|a?bb(dcc*|(d|c)))ab(c|d)*(ccd)?",
        ]

        for regex_str in regex_strings:
            nfa_1 = NFA.from_regex(regex_str)
            gnfa_1 = GNFA.from_nfa(nfa_1)
            regex_1 = gnfa_1.to_regex()
            nfa_2 = NFA.from_regex(regex_1)

            # Test equality under NFA regex conversion
            self.assertEqual(nfa_1, nfa_2)

            dfa_1 = DFA.from_nfa(nfa_1)
            gnfa_2 = GNFA.from_dfa(dfa_1)
            regex_2 = gnfa_2.to_regex()
            dfa_2 = DFA.from_nfa(NFA.from_regex(regex_2))

            # Test equality through DFA regex conversion
            self.assertEqual(dfa_1, dfa_2)

    def test_show_diagram(self) -> None:
        """
        Should construct the diagram for a GNFA.
        """

        graph = self.gnfa.show_diagram()

        node_names = {node.get_name() for node in graph.nodes()}
        self.assertTrue(set(self.gnfa.states).issubset(node_names))
        self.assertEqual(len(self.gnfa.states) + 1, len(node_names))

        for state in self.dfa.states:
            node = graph.get_node(state)
            expected_shape = (
                "doublecircle" if state in self.gnfa.final_states else "circle"
            )
            self.assertEqual(node.attr["shape"], expected_shape)

        expected_transitions = {
            ("q_in", "ε", "q0"),
            ("q1", "ε", "q2"),
            ("q1", "ε", "q_f"),
            ("q1", "a", "q1"),
            ("q2", "b", "q0"),
            ("q0", "a", "q1"),
        }
        seen_transitions = {
            (edge[0], edge.attr["label"], edge[1]) for edge in graph.edges()
        }

        self.assertTrue(expected_transitions.issubset(seen_transitions))
        self.assertEqual(len(expected_transitions) + 1, len(seen_transitions))

        source, symbol, dest = list(seen_transitions - expected_transitions)[0]
        self.assertEqual(symbol, "")
        self.assertEqual(dest, self.gnfa.initial_state)
        self.assertTrue(source not in self.gnfa.states)

    def test_show_diagram_write_file(self) -> None:
        """
        Should construct the diagram for a NFA
        and write it to the specified file.
        """
        diagram_path = os.path.join(self.temp_dir_path, "test_gnfa.png")
        try:
            os.remove(diagram_path)
        except OSError:
            pass
        self.assertFalse(os.path.exists(diagram_path))
        self.gnfa.show_diagram(path=diagram_path)
        self.assertTrue(os.path.exists(diagram_path))
        os.remove(diagram_path)



"""Classes and functions for testing the behavior of MNTMs."""

import math
import random
import types
from typing import Set, cast
from unittest.mock import MagicMock, patch

from frozendict import frozendict

import automata.base.exceptions as exceptions
import automata.tm.exceptions as tm_exceptions
import tests.test_tm as test_tm
from automata.tm.mntm import MNTM


class TestMNTM(test_tm.TestTM):
    """A test class for testing multitape nondeterministic Turing machines."""

    def test_init_mntm(self) -> None:
        """Should copy MNTM if passed into MNTM constructor."""
        new_mntm = self.mntm1.copy()
        self.assertIsNot(new_mntm, self.mntm1)

    def test_init_mntm_missing_formal_params(self) -> None:
        """Should raise an error if formal MNTM parameters are missing."""
        with self.assertRaises(TypeError):
            MNTM(  # type: ignore
                states={"q0", "q1", "q2", "q3", "q4"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "x", "y", "."},
                initial_state="q0",
                blank_symbol=".",
                final_states={"q4"},
            )

    @patch("automata.tm.mntm.MNTM.validate")
    def test_init_validation(self, validate: MagicMock) -> None:
        """Should validate MNTM when initialized."""
        self.mntm1.copy()
        validate.assert_called_once_with()

    def test_copy_ntm(self) -> None:
        """Should create exact copy of MNTM if copy() method is called."""
        new_mntm = self.mntm1.copy()
        self.assertIsNot(new_mntm, self.mntm1)

    def test_mntm_immutable_attr_set(self) -> None:
        with self.assertRaises(AttributeError):
            self.mntm1.states = set()

    def test_mntm_immutable_attr_del(self) -> None:
        with self.assertRaises(AttributeError):
            del self.mntm1.states

    def test_mntm_immutable_dict(self) -> None:
        """Should create a DPDA whose contents are fully immutable/hashable"""
        self.assertIsInstance(hash(frozendict(self.mntm1.input_parameters)), int)

    def test_validate_input_symbol_subset(self) -> None:
        """Should raise error if input symbols are not a strict superset of tape
        symbols."""
        with self.assertRaises(exceptions.MissingSymbolError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_transition_state(self) -> None:
        """Should raise error if a transition state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q5": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_transition_symbol(self) -> None:
        """Should raise error if a transition symbol is invalid."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("5", "#"): [("q1", (("#", "R"), ("#", "R")))],
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_transition_result_state(self) -> None:
        """Should raise error if a transition result state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q5", (("#", "L"), ("#", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_transition_result_symbol(self) -> None:
        """Should raise error if a transition result symbol is invalid."""
        with self.assertRaises(exceptions.InvalidSymbolError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q1", ((".", "L"), ("#", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_transition_result_direction(self) -> None:
        """Should raise error if a transition result direction is invalid."""
        with self.assertRaises(tm_exceptions.InvalidDirectionError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q1", (("#", "U"), ("#", "R")))],  # type: ignore
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_invalid_initial_state(self) -> None:
        """Should raise error if the initial state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q5",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_initial_state_transitions(self) -> None:
        """Should raise error if the initial state has no transitions."""
        with self.assertRaises(exceptions.MissingStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q1": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_nonfinal_initial_state(self) -> None:
        """Should raise error if the initial state is a final state."""
        with self.assertRaises(exceptions.InitialStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q0", "q1"},
            )

    def test_validate_invalid_final_state(self) -> None:
        """Should raise error if the final state is invalid."""
        with self.assertRaises(exceptions.InvalidStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q5"},
            )

    def test_validate_invalid_final_state_non_str(self) -> None:
        """Should raise InvalidStateError even for non-string final states."""
        with self.assertRaises(exceptions.InvalidStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={5},
            )

    def test_validate_final_state_transitions(self) -> None:
        """Should raise error if a final state has any transitions."""
        with self.assertRaises(exceptions.FinalStateError):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    },
                    "q1": {("0", "#"): [("q0", (("0", "L"), ("0", "R")))]},
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_tapes_consistency_too_few_specified(self) -> None:
        with self.assertRaises(tm_exceptions.InconsistentTapesException):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=3,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_validate_tapes_consistency_too_many_specified(self) -> None:
        with self.assertRaises(tm_exceptions.InconsistentTapesException):
            MNTM(
                states={"q0", "q1"},
                input_symbols={"0", "1"},
                tape_symbols={"0", "1", "#"},
                n_tapes=2,
                transitions={
                    "q0": {
                        ("1", "#"): [("q0", (("1", "R"), ("1", "R")))],
                        ("0", "#"): [("q0", (("0", "R"), ("#", "N"), ("#", "R")))],
                        ("#", "#"): [("q1", (("#", "N"), ("#", "N")))],
                    }
                },
                initial_state="q0",
                blank_symbol="#",
                final_states={"q1"},
            )

    def test_read_extended_tape(self) -> None:
        self.assertEqual(
            self.mntm1._read_extended_tape("10^10_1^00_00#^_", "^"), ("0", "1", "#")
        )
        self.assertEqual(
            self.mntm1._read_extended_tape("1.10_1.00_0.#_", "."), ("1", "1", "0")
        )
        self.assertEqual(
            self.mntm1._read_extended_tape("10#^_00#^_00^_", "^"), ("#", "#", "0")
        )

        with self.assertRaisesRegex(
            tm_exceptions.MalformedExtendedTapeError,
            "head symbol was found on leftmost " + "end of the extended tape",
        ):
            self.assertEqual(
                self.mntm1._read_extended_tape("^10#_1^010#_00^", "^"), ("", "1", "0")
            )

        with self.assertRaisesRegex(
            tm_exceptions.MalformedExtendedTapeError,
            "no head symbol found on one of the " + "virtual tapes",
        ):
            self.assertEqual(
                self.mntm1._read_extended_tape("0^10#_1010#_00^_", "^"), ("0", "", "0")
            )

        with self.assertRaisesRegex(
            tm_exceptions.MalformedExtendedTapeError,
            "there must be 1 virtual head for " + "every tape separator symbol",
        ):
            self.assertEqual(
                self.mntm1._read_extended_tape("0^1010^10#^", "^"), ("0", "0", "#")
            )

        with self.assertRaisesRegex(
            tm_exceptions.MalformedExtendedTapeError,
            "more than one head symbol found on " + "one of the virtual tapes",
        ):
            self.assertEqual(
                self.mntm1._read_extended_tape("0^101010^_#^_", "^"), ("0", "0", "#")
            )

    def test_read_input_as_ntm(self) -> None:
        validation_generator = self.mntm2.read_input_as_ntm("#0000")
        configs = list(validation_generator)
        first_config = set(configs[0]).pop()
        self.assertEqual(first_config.state, "q-1")
        self.assertEqual(str(first_config.tape), "TMTape('#^0000_#^_#^_', '#', 0)")
        last_config = set(configs[-1]).pop()
        self.assertEqual(last_config.state, "qf")
        self.assertEqual(
            str(last_config.tape), "TMTape('#0000#^_#0000#^_#XYYY#^_', '#', 23)"
        )

        # Test that it throws a RejectionException
        with self.assertRaises(exceptions.RejectionException):
            for _ in self.mntm2.read_input_as_ntm("#00"):
                pass

        # Thoroughly test the MNTM as an NTM and check that the final tapes are the same
        # regardless of whether the MNTM is run as an NTM or an MNTM
        mntms = [self.mntm1, self.mntm2, self.mntm3]
        inputs = ["0110", "#0000", "0101"]

        for mntm, input_str in zip(mntms, inputs):
            # Test that it doesn't throw a RejectionException
            self.assertTrue(mntm.accepts_input(input_str))
            configs = list(mntm.read_input_stepwise(input_str))
            final_mntm_config = cast(Set, configs[-1]).pop()

            # Test that the final tape of the MNTM is the same as the final tape of the
            # MNTM ran as an NTM.
            # First, join the tapes of the MNTM, as they would (and should) be in the
            # NTM version ^ being the head symbol, _ being the tape separator
            final_mntm_tape = ""
            for tape in final_mntm_config.tapes:
                curr_tape = "".join(tape.tape)

                # Add a ^ after the current position, for the head symbol
                curr_tape_pos = tape.current_position + 1

                # For self.mntm1 with input 0110, the final tape of the MNTM should be
                # 0110#^_11#^_ in the end as a result of concatenating:
                # MTMConfig... ('q1', (TMTape('0110#', '#', 4), TMTape('11#', '#', 2)))
                final_mntm_tape += (
                    curr_tape[:curr_tape_pos] + "^" + curr_tape[curr_tape_pos:] + "_"
                )

            # Now, run the MNTM as an NTM
            configs = list(mntm.read_input_as_ntm(input_str))
            final_ntm_config = cast(Set, configs[-1]).pop()

            # e.g., 0110#^_11#^_ for self.mntm1 with input 0110 as a result of:
            # TMConfiguration('q1', TMTape('0110#^_11#^_', '#', 11))
            final_ntm_tape = "".join(final_ntm_config.tape.tape)

            # Finally, assert that the final tapes are the same
            self.assertEqual(final_mntm_tape, final_ntm_tape)

    def test_read_input_accepted(self) -> None:
        """Should return correct state if acceptable TM input is given."""
        final_config = self.mntm1.read_input("0101101011").pop()
        self.assertEqual(final_config.state, "q1")
        self.assertEqual(str(final_config.tapes[0]), "TMTape('0101101011#', '#', 10)")
        self.assertEqual(str(final_config.tapes[1]), "TMTape('111111#', '#', 6)")

    def test_read_input_step(self) -> None:
        """Should return validation generator if step flag is supplied."""
        validation_generator = self.mntm1.read_input_stepwise("0010101111")
        self.assertIsInstance(validation_generator, types.GeneratorType)
        configs = list(validation_generator)
        first_config = configs[0].pop()
        self.assertEqual(first_config.state, "q0")
        self.assertEqual(str(first_config.tapes[0]), "TMTape('0010101111', '#', 0)")
        self.assertEqual(str(first_config.tapes[1]), "TMTape('#', '#', 0)")
        last_config = configs[-1].pop()
        self.assertEqual(last_config.state, "q1")
        self.assertEqual(str(last_config.tapes[0]), "TMTape('0010101111#', '#', 10)")
        self.assertEqual(str(last_config.tapes[1]), "TMTape('111111#', '#', 6)")

    def test_read_input_rejection(self) -> None:
        """Should raise error if the machine halts."""
        with self.assertRaises(exceptions.RejectionException):
            self.mntm1.read_input("2")

    def test_read_input_rejection_invalid_symbol(self) -> None:
        """Should raise error if an invalid symbol is read."""
        with self.assertRaises(exceptions.RejectionException):
            self.mntm2.read_input("1")

    def test_accepts_input_true(self) -> None:
        """Should return False if MNTM input is not accepted."""
        test_limit = 20
        for i in range(test_limit):
            input_str_1 = ""
            input_str_2 = "#0"
            input_str_3 = "1"
            for _ in range(i):
                k = random.randint(0, 1)
                input_str_1 += str(k)
                input_str_2 += "0"
                input_str_3 += str(k)

            input_str_3 += input_str_3
            # Should accept all
            self.assertTrue(self.mntm1.accepts_input(input_str_1))
            self.assertTrue(self.mntm3.accepts_input(input_str_3))

            # Should not accept because this would not be of the form ww
            self.assertFalse(
                self.mntm3.accepts_input(input_str_3 + str(random.randint(0, 1)))
            )

            # Should accept only if input string's length is a perfect square
            if self.is_perfect_square(len(input_str_2) - 1):
                self.assertTrue(self.mntm2.accepts_input(input_str_2))
            else:
                self.assertFalse(self.mntm2.accepts_input(input_str_2))

    def test_accepts_input_false(self) -> None:
        """Should return False if MNTM input is rejected."""
        self.assertFalse(self.mntm1.accepts_input("000012"))
        self.assertFalse(self.mntm2.accepts_input("#00000"))

    @staticmethod
    def is_perfect_square(number: int) -> bool:
        return number == int(math.sqrt(number)) ** 2



"""Classes and functions for testing the behavior of both DFAs and NFAs."""

import unittest

from automata.fa.dfa import DFA
from automata.fa.fa import FA
from automata.fa.gnfa import GNFA
from automata.fa.nfa import NFA


class TestFA(unittest.TestCase):
    """A test class for testing all finite automata."""

    dfa: DFA
    nfa: NFA
    gnfa: GNFA

    def setUp(self) -> None:
        """Reset test automata before every test function."""
        # DFA which matches all binary strings ending in an odd number of '1's
        self.dfa = DFA(
            states={"q0", "q1", "q2"},
            input_symbols={"0", "1"},
            transitions={
                "q0": {"0": "q0", "1": "q1"},
                "q1": {"0": "q0", "1": "q2"},
                "q2": {"0": "q2", "1": "q1"},
            },
            initial_state="q0",
            final_states={"q1"},
        )
        # NFA which matches strings beginning with 'a', ending with 'a', and
        # containing no consecutive 'b's
        self.nfa = NFA(
            states={"q0", "q1", "q2"},
            input_symbols={"a", "b"},
            transitions={
                "q0": {"a": {"q1"}},
                "q1": {"a": {"q1"}, "": {"q2"}},
                "q2": {"b": {"q0"}},
            },
            initial_state="q0",
            final_states={"q1"},
        )
        # GNFA which matches strings beginning with 'a', ending with 'a', and containing
        # no consecutive 'b's
        self.gnfa = GNFA(
            states={"q_in", "q_f", "q0", "q1", "q2"},
            input_symbols={"a", "b"},
            transitions={
                "q0": {"q1": "a", "q_f": None, "q2": None, "q0": None},
                "q1": {"q1": "a", "q2": "", "q_f": "", "q0": None},
                "q2": {"q0": "b", "q_f": None, "q2": None, "q1": None},
                "q_in": {"q0": "", "q_f": None, "q2": None, "q1": None},
            },
            initial_state="q_in",
            final_state="q_f",
        )


class TestFAAbstract(unittest.TestCase):
    def test_abstract_methods_not_implemented(self) -> None:
        """Should raise NotImplementedError when calling abstract methods."""

        abstract_methods = {
            "iter_transitions": (FA,),
            "_get_input_path": (FA, ""),
        }
        for method_name, method_args in abstract_methods.items():
            with self.assertRaises(NotImplementedError):
                getattr(FA, method_name)(*method_args)

        with self.assertRaises(NotImplementedError):
            getattr(FA, "_get_input_path")(FA, "")

    def test_get_state_name(self) -> None:
        """Tests get_state_name function."""

        self.assertEqual(FA._get_state_name(""), "λ")
        self.assertEqual(FA._get_state_name("abc"), "abc")
        self.assertEqual(FA._get_state_name(123), "123")
        self.assertEqual(FA._get_state_name(frozenset()), "∅")

        original_sets = [frozenset("abc"), frozenset("dcgf")]

        for original_set in original_sets:
            set_output = frozenset(
                FA._get_state_name(original_set)
                .replace("{", "")
                .replace("}", "")
                .replace("'", "")
                .replace(" ", "")
                .split(",")
            )
            self.assertEqual(set_output, original_set)

        original_tuple = tuple("abc")

        tuple_output = tuple(
            FA._get_state_name(original_tuple)
            .replace("(", "")
            .replace(")", "")
            .replace("'", "")
            .replace(" ", "")
            .split(",")
        )

        self.assertEqual(tuple_output, original_tuple)



"""Exception classes shared by all automata."""

from dataclasses import dataclass


class AutomatonException(Exception):
    """The base class for all automaton-related errors."""

    pass


class InvalidStateError(AutomatonException):
    """A state is not a valid state for this automaton."""

    pass


class InvalidSymbolError(AutomatonException):
    """A symbol is not a valid symbol for this automaton."""

    pass


class MissingStateError(AutomatonException):
    """A state is missing from the automaton definition."""

    pass


class MissingSymbolError(AutomatonException):
    """A symbol is missing from the automaton definition."""

    pass


class InitialStateError(AutomatonException):
    """The initial state fails to meet some required condition."""

    pass


class FinalStateError(AutomatonException):
    """A final state fails to meet some required condition."""

    pass


class RejectionException(AutomatonException):
    """The input was rejected by the automaton."""

    pass


class RegexException(Exception):
    """The base class for all regular expression related errors"""

    pass


@dataclass
class LexerError(RegexException):
    """An exception raised for issues in lexing"""

    message: str
    position: int


class InvalidRegexError(RegexException):
    """Regular expression is invalid"""

    pass


class SymbolMismatchError(AutomatonException):
    """The input symbols between the given automata do not match"""

    pass


class EmptyLanguageException(AutomatonException):
    """The operation cannot be performed because the language is empty"""

    pass


class InfiniteLanguageException(AutomatonException):
    """The operation cannot be performed because the language is infinite"""

    pass


class DiagramException(AutomatonException):
    """The diagram cannot be produced"""

    pass
